#!/usr/bin/env python3
import unittest
import numpy as np
import ahrs

# Often used constants
SQ22 = np.sqrt(2)/2
COS_SQ22 = np.cos(SQ22)
SIN_SQ22 = np.sin(SQ22)
EXP_SQ22 = np.exp(SQ22)

class TestQuaternion(unittest.TestCase):
    def setUp(self) -> None:
        # Identity Quaternion
        self.q0 = ahrs.Quaternion()
        # Versor from 3D array
        self.vector3 = np.random.random(3)-0.5
        self.q3 = ahrs.Quaternion(self.vector3)
        self.vector3 /= np.linalg.norm(self.vector3)
        # Versor from 4D array
        self.vector4 = np.random.random(4)-0.5
        self.q4 = ahrs.Quaternion(self.vector4)
        self.vector4 /= np.linalg.norm(self.vector4)
        self.decimal_precision = 15

    def test_identity_quaternion(self):
        self.assertEqual(self.q0.w, 1.0)
        self.assertEqual(self.q0.x, 0.0)
        self.assertEqual(self.q0.y, 0.0)
        self.assertEqual(self.q0.z, 0.0)
        np.testing.assert_equal(self.q0.v, np.zeros(3))

    def test_conjugate(self):
        qc = self.q4.conjugate
        self.assertAlmostEqual(qc[0], self.q4.w, places=self.decimal_precision)
        self.assertAlmostEqual(qc[1], -self.q4.x, places=self.decimal_precision)
        self.assertAlmostEqual(qc[2], -self.q4.y, places=self.decimal_precision)
        self.assertAlmostEqual(qc[3], -self.q4.z, places=self.decimal_precision)
        np.testing.assert_almost_equal(qc[1:], -self.q4.v, decimal=self.decimal_precision)

    def test_conj(self):
        qc = self.q4.conj
        self.assertAlmostEqual(qc[0], self.q4.w, places=self.decimal_precision)
        self.assertAlmostEqual(qc[1], -self.q4.x, places=self.decimal_precision)
        self.assertAlmostEqual(qc[2], -self.q4.y, places=self.decimal_precision)
        self.assertAlmostEqual(qc[3], -self.q4.z, places=self.decimal_precision)
        np.testing.assert_almost_equal(qc[1:], -self.q4.v, decimal=self.decimal_precision)

    def test_inverse(self):
        qi = self.q4.inverse
        self.assertAlmostEqual(qi[0], self.q4.w, places=self.decimal_precision)
        self.assertAlmostEqual(qi[1], -self.q4.x, places=self.decimal_precision)
        self.assertAlmostEqual(qi[2], -self.q4.y, places=self.decimal_precision)
        self.assertAlmostEqual(qi[3], -self.q4.z, places=self.decimal_precision)
        np.testing.assert_almost_equal(qi[1:], -self.q4.v, decimal=self.decimal_precision)
        # Not versor
        qn = ahrs.Quaternion([1.0, 2.0, 3.0, 4.0], versor=False)
        expected_inverse = np.array([1.0, -2.0, -3.0, -4.0])/np.linalg.norm([1.0, 2.0, 3.0, 4.0])
        np.testing.assert_almost_equal(qn.inverse, expected_inverse, decimal=self.decimal_precision)

    def test_inv(self):
        qi = self.q4.inv
        self.assertAlmostEqual(qi[0], self.q4.w, places=self.decimal_precision)
        self.assertAlmostEqual(qi[1], -self.q4.x, places=self.decimal_precision)
        self.assertAlmostEqual(qi[2], -self.q4.y, places=self.decimal_precision)
        self.assertAlmostEqual(qi[3], -self.q4.z, places=self.decimal_precision)
        np.testing.assert_almost_equal(qi[1:], -self.q4.v, decimal=self.decimal_precision)
        # Not versor
        qn = ahrs.Quaternion([1.0, 2.0, 3.0, 4.0], versor=False)
        expected_inverse = np.array([1.0, -2.0, -3.0, -4.0])/np.linalg.norm([1.0, 2.0, 3.0, 4.0])
        np.testing.assert_almost_equal(qn.inv, expected_inverse, decimal=self.decimal_precision)

    def test_is_pure(self):
        self.assertFalse(self.q0.is_pure())
        self.assertTrue(self.q3.is_pure())
        self.assertFalse(self.q4.is_pure())

    def test_is_real(self):
        self.assertTrue(self.q0.is_real())
        self.assertFalse(self.q3.is_real())
        self.assertFalse(self.q4.is_real())

    def test_is_versor(self):
        self.assertTrue(self.q0.is_versor())
        self.assertTrue(self.q3.is_versor())
        self.assertTrue(self.q4.is_versor())

    def test_is_identity(self):
        self.assertTrue(self.q0.is_identity())
        self.assertFalse(self.q3.is_identity())
        self.assertFalse(self.q4.is_identity())

    def test_versor_from_3d_array(self):
        self.assertAlmostEqual(self.q3.w, 0.0, places=self.decimal_precision)
        self.assertAlmostEqual(self.q3.x, self.vector3[0], places=self.decimal_precision)
        self.assertAlmostEqual(self.q3.y, self.vector3[1], places=self.decimal_precision)
        self.assertAlmostEqual(self.q3.z, self.vector3[2], places=self.decimal_precision)
        np.testing.assert_almost_equal(self.q3.v, self.vector3, decimal=self.decimal_precision)

    def test_versor_from_4d_array(self):
        self.assertAlmostEqual(self.q4.w, self.vector4[0], places=self.decimal_precision)
        self.assertAlmostEqual(self.q4.x, self.vector4[1], places=self.decimal_precision)
        self.assertAlmostEqual(self.q4.y, self.vector4[2], places=self.decimal_precision)
        self.assertAlmostEqual(self.q4.z, self.vector4[3], places=self.decimal_precision)
        np.testing.assert_almost_equal(self.q4.v, self.vector4[1:], decimal=self.decimal_precision)

    def test_wrong_input_array(self):
        self.assertRaises(TypeError, ahrs.Quaternion, True)
        self.assertRaises(TypeError, ahrs.Quaternion, 3.0)
        self.assertRaises(TypeError, ahrs.Quaternion, "[1.0, 2.0, 3.0, 4.0]")
        self.assertRaises(ValueError, ahrs.Quaternion, np.random.random((6, 3)))
        self.assertRaises(ValueError, ahrs.Quaternion, np.random.random((6, 4)))
        self.assertRaises(ValueError, ahrs.Quaternion, np.random.random(2))
        self.assertRaises(ValueError, ahrs.Quaternion, np.random.random(2).tolist())
        self.assertRaises(ValueError, ahrs.Quaternion, np.zeros(3))

    def test_wrong_input_dcm(self):
        self.assertRaises(TypeError, ahrs.Quaternion, dcm=3)
        self.assertRaises(TypeError, ahrs.Quaternion, dcm=3.0)
        self.assertRaises(TypeError, ahrs.Quaternion, dcm="np.identity(3)")
        self.assertRaises(ValueError, ahrs.Quaternion, dcm=np.random.random((3, 3)))
        self.assertRaises(ValueError, ahrs.Quaternion, dcm=-np.identity(3))

    def test_random_attitudes(self):
        q = ahrs.Quaternion(random=True)
        self.assertEqual(q.shape, (4,))
        self.assertTrue(np.all(np.abs(np.linalg.norm(q)-1) < 1e-15))

    def test_rpy(self):
        np.testing.assert_almost_equal(ahrs.Quaternion(rpy=[np.pi, 0, 0]).to_array(), [0.0, 1.0, 0.0, 0.0], decimal=self.decimal_precision)
        np.testing.assert_almost_equal(ahrs.Quaternion(rpy=[0, np.pi, 0]).to_array(), [0.0, 0.0, 1.0, 0.0], decimal=self.decimal_precision)
        np.testing.assert_almost_equal(ahrs.Quaternion(rpy=[0, 0, np.pi]).to_array(), [0.0, 0.0, 0.0, 1.0], decimal=self.decimal_precision)
        np.testing.assert_almost_equal(ahrs.Quaternion(rpy=[np.pi/2, 0, 0]).to_array(), [SQ22, SQ22, 0.0, 0.0], decimal=self.decimal_precision)
        np.testing.assert_almost_equal(ahrs.Quaternion(rpy=[0, np.pi/2, 0]).to_array(), [SQ22, 0.0, SQ22, 0.0], decimal=self.decimal_precision)
        np.testing.assert_almost_equal(ahrs.Quaternion(rpy=[0, 0, np.pi/2]).to_array(), [SQ22, 0.0, 0.0, SQ22], decimal=self.decimal_precision)
        # Older calls using 'angles' instead of 'rpy'. Just for compatibility
        np.testing.assert_almost_equal(ahrs.Quaternion(angles=[np.pi/2, 0, 0]).to_array(), [SQ22, SQ22, 0.0, 0.0], decimal=self.decimal_precision)
        np.testing.assert_almost_equal(ahrs.Quaternion(angles=[0, np.pi/2, 0]).to_array(), [SQ22, 0.0, SQ22, 0.0], decimal=self.decimal_precision)
        np.testing.assert_almost_equal(ahrs.Quaternion(angles=[0, 0, np.pi/2]).to_array(), [SQ22, 0.0, 0.0, SQ22], decimal=self.decimal_precision)

    def test_exponential(self):
        np.testing.assert_allclose(ahrs.Quaternion([1, 0, 0, 0]).exponential, [1, 0, 0, 0])
        np.testing.assert_allclose(ahrs.Quaternion([0, 1, 0, 0]).exponential, [np.cos(1), np.sin(1), 0, 0])
        np.testing.assert_allclose(ahrs.Quaternion([0, 0, 1, 0]).exponential, [np.cos(1), 0, np.sin(1), 0])
        np.testing.assert_allclose(ahrs.Quaternion([0, 0, 0, 1]).exponential, [np.cos(1), 0, 0, np.sin(1)])
        np.testing.assert_allclose(ahrs.Quaternion([1, 1, 0, 0]).exponential, np.array([COS_SQ22, SIN_SQ22, 0, 0])*EXP_SQ22)
        np.testing.assert_allclose(ahrs.Quaternion([1, 0, 1, 0]).exponential, np.array([COS_SQ22, 0, SIN_SQ22, 0])*EXP_SQ22)
        np.testing.assert_allclose(ahrs.Quaternion([1, 0, 0, 1]).exponential, np.array([COS_SQ22, 0, 0, SIN_SQ22])*EXP_SQ22)

    def test_exp(self):
        np.testing.assert_allclose(ahrs.Quaternion([1, 0, 0, 0]).exp, [1, 0, 0, 0])
        np.testing.assert_allclose(ahrs.Quaternion([0, 1, 0, 0]).exp, [np.cos(1), np.sin(1), 0, 0])
        np.testing.assert_allclose(ahrs.Quaternion([0, 0, 1, 0]).exp, [np.cos(1), 0, np.sin(1), 0])
        np.testing.assert_allclose(ahrs.Quaternion([0, 0, 0, 1]).exp, [np.cos(1), 0, 0, np.sin(1)])
        np.testing.assert_allclose(ahrs.Quaternion([1, 1, 0, 0]).exp, np.array([COS_SQ22, SIN_SQ22, 0, 0])*EXP_SQ22)
        np.testing.assert_allclose(ahrs.Quaternion([1, 0, 1, 0]).exp, np.array([COS_SQ22, 0, SIN_SQ22, 0])*EXP_SQ22)
        np.testing.assert_allclose(ahrs.Quaternion([1, 0, 0, 1]).exp, np.array([COS_SQ22, 0, 0, SIN_SQ22])*EXP_SQ22)

    def test_logarithm(self):
        np.testing.assert_allclose(ahrs.Quaternion([1, 0, 0, 0]).logarithm, [0, 0, 0, 0])
        np.testing.assert_allclose(ahrs.Quaternion([0, 1, 0, 0]).logarithm, [0, np.pi/2, 0, 0])
        np.testing.assert_allclose(ahrs.Quaternion([0, 0, 1, 0]).logarithm, [0, 0, np.pi/2, 0])
        np.testing.assert_allclose(ahrs.Quaternion([0, 0, 0, 1]).logarithm, [0, 0, 0, np.pi/2])
        np.testing.assert_allclose(ahrs.Quaternion([1, 1, 0, 0]).logarithm, [0, np.pi/4, 0, 0])
        np.testing.assert_allclose(ahrs.Quaternion([1, 0, 1, 0]).logarithm, [0, 0, np.pi/4, 0])
        np.testing.assert_allclose(ahrs.Quaternion([1, 0, 0, 1]).logarithm, [0, 0, 0, np.pi/4])

    def test_log(self):
        np.testing.assert_allclose(ahrs.Quaternion([1, 0, 0, 0]).log, [0, 0, 0, 0])
        np.testing.assert_allclose(ahrs.Quaternion([0, 1, 0, 0]).log, [0, np.pi/2, 0, 0])
        np.testing.assert_allclose(ahrs.Quaternion([0, 0, 1, 0]).log, [0, 0, np.pi/2, 0])
        np.testing.assert_allclose(ahrs.Quaternion([0, 0, 0, 1]).log, [0, 0, 0, np.pi/2])
        np.testing.assert_allclose(ahrs.Quaternion([1, 1, 0, 0]).log, [0, np.pi/4, 0, 0])
        np.testing.assert_allclose(ahrs.Quaternion([1, 0, 1, 0]).log, [0, 0, np.pi/4, 0])
        np.testing.assert_allclose(ahrs.Quaternion([1, 0, 0, 1]).log, [0, 0, 0, np.pi/4])

    def test_str(self):
        self.assertEqual(str(self.q0), "(1.0000 +0.0000i +0.0000j +0.0000k)")
        self.assertEqual(str(self.q3), f"(0.0000 {self.vector3[0]:+.4f}i {self.vector3[1]:+.4f}j {self.vector3[2]:+.4f}k)")
        self.assertEqual(str(self.q4), f"({self.vector4[0]:-.4f} {self.vector4[1]:+.4f}i {self.vector4[2]:+.4f}j {self.vector4[3]:+.4f}k)")
        self.assertEqual(str(ahrs.Quaternion([0, 0, 0, 1], order='S')), "(0.0000i +0.0000j +0.0000k +1.0000)")
        self.assertEqual(str(ahrs.Quaternion([1, 0, 0, 1], order='S')), "(0.7071i +0.0000j +0.0000k +0.7071)")

    def test_addition(self):
        p = ahrs.Quaternion([0.0, 1.0, 0.0, 0.0])
        q = ahrs.Quaternion([0.0, 0.0, 1.0, 0.0])
        r = ahrs.Quaternion([0.0, 0.0, 0.0, 1.0])
        np.testing.assert_allclose((p+p).to_array(), [0.0, 1.0, 0.0, 0.0])
        np.testing.assert_allclose((p+q).to_array(), [0.0, SQ22, SQ22, 0.0])
        np.testing.assert_allclose((q+r).to_array(), [0.0, 0.0, SQ22, SQ22])
        np.testing.assert_allclose((r+p).to_array(), [0.0, SQ22, 0.0, SQ22])

    def test_subtraction(self):
        p = ahrs.Quaternion([0.0, 1.0, 0.0, 0.0])
        q = ahrs.Quaternion([0.0, 0.0, 1.0, 0.0])
        r = ahrs.Quaternion([0.0, 0.0, 0.0, 1.0])
        np.testing.assert_allclose((p-q).to_array(), [0.0, SQ22, -SQ22, 0.0])
        np.testing.assert_allclose((q-r).to_array(), [0.0, 0.0, SQ22, -SQ22])
        np.testing.assert_allclose((r-p).to_array(), [0.0, -SQ22, 0.0, SQ22])

    def test_multiplication(self):
        p = ahrs.Quaternion([1., 0., 0., 0.])
        q = ahrs.Quaternion([1., 0., 1., 0.])
        r = ahrs.Quaternion([0., 0., 1., 0.])
        np.testing.assert_allclose(p*q, [SQ22, 0, SQ22, 0])
        np.testing.assert_allclose(p*r, [0, 0, 1, 0])
        np.testing.assert_allclose(q*r, [-SQ22, 0, SQ22, 0])
        np.testing.assert_allclose(p*np.array([0., 0., 1., 0.]), [0, 0, 1, 0])
        np.testing.assert_allclose(q*np.array([0., 0., 1., 0.]), [-SQ22, 0, SQ22, 0])
        # Non-unitary vector (non-versor)
        np.testing.assert_allclose(p*np.array([1., 0., 1., 0.]), [1, 0, 1, 0])
        np.testing.assert_allclose(q*np.array([1., 0., 1., 0.]), [0, 0, np.sqrt(2), 0])
        np.testing.assert_allclose(r*np.array([1., 0., 1., 0.]), [-1, 0, 1, 0])

    def test_matrix_multiplication(self):
        p = ahrs.Quaternion([1., 0., 0., 0.])
        q = ahrs.Quaternion([1., 0., 1., 0.])
        r = ahrs.Quaternion([0., 0., 1., 0.])
        np.testing.assert_allclose(p @ q, np.array([SQ22, 0, SQ22, 0]))
        np.testing.assert_allclose(p @ r, np.array([0, 0, 1, 0]))
        np.testing.assert_allclose(q @ r, np.array([-SQ22, 0, SQ22, 0]))
        np.testing.assert_allclose(p @ np.array([0., 0., 1., 0.]), np.array([0, 0, 1, 0]))
        np.testing.assert_allclose(q @ np.array([0., 0., 1., 0.]), np.array([-SQ22, 0, SQ22, 0]))
        # Non-unitary vector (non-versor)
        np.testing.assert_allclose(p @ np.array([1., 0., 1., 0.]), np.array([1, 0, 1, 0]))
        np.testing.assert_allclose(q @ np.array([1., 0., 1., 0.]), np.array([0, 0, np.sqrt(2), 0]))
        np.testing.assert_allclose(r @ np.array([1., 0., 1., 0.]), np.array([-1, 0, 1, 0]))

    def test_normalize(self):
        q = ahrs.Quaternion([1., 0., 1., 0.], versor=False)
        q.normalize()
        np.testing.assert_allclose(q.to_array(), [SQ22, 0, SQ22, 0])

    def test_mult_L(self):
        p = ahrs.Quaternion([1., 0., 0., 0.])
        q = ahrs.Quaternion([1., 0., 1., 0.])
        r = ahrs.Quaternion([0., 0., 1., 0.])
        skewed_matrix = np.array([[0, 0, -1, 0], [0, 0, 0, 1], [1, 0, 0, 0], [0, -1, 0, 0]])
        np.testing.assert_allclose(p.mult_L(), np.identity(4))
        np.testing.assert_allclose(q.mult_L(), SQ22 * (skewed_matrix + np.identity(4)))
        np.testing.assert_allclose(r.mult_L(), skewed_matrix)

    def test_mult_R(self):
        p = ahrs.Quaternion([1., 0., 0., 0.])
        q = ahrs.Quaternion([1., 0., 1., 0.])
        r = ahrs.Quaternion([0., 0., 1., 0.])
        skewed_matrix = np.array([[0, 0, -1, 0], [0, 0, 0, -1], [1, 0, 0, 0], [0, 1, 0, 0]])
        np.testing.assert_allclose(p.mult_R(), np.identity(4))
        np.testing.assert_allclose(q.mult_R(), SQ22 * (skewed_matrix + np.identity(4)))
        np.testing.assert_allclose(r.mult_R(), skewed_matrix)

    def test_rotate(self):
        q = ahrs.Quaternion([1., 0., 0., 0.])
        np.testing.assert_allclose(q.rotate([1., 0., 0.]), [1., 0., 0.])
        np.testing.assert_allclose(q.rotate([0., 1., 0.]), [0., 1., 0.])
        np.testing.assert_allclose(q.rotate([1., 2., 3.]), [1., 2., 3.])
        q = ahrs.Quaternion([1., 0., 1., 0.])
        np.testing.assert_allclose(q.rotate([1., 0., 0.]), [0., 0., -1.], atol=self.decimal_precision)
        np.testing.assert_allclose(q.rotate([0., 1., 0.]), [0., 1., 0.])
        np.testing.assert_allclose(q.rotate([1., 2., 3.]), [3., 2., -1.])
        q = ahrs.Quaternion([0., 0., 1., 0.])
        np.testing.assert_allclose(q.rotate([1., 0., 0.]), [-1., 0., 0.])
        np.testing.assert_allclose(q.rotate([0., 1., 0.]), [0., 1., 0.])
        np.testing.assert_allclose(q.rotate([1., 2., 3.]), [-1., 2., -3.])

    def test_rotate_2d_array(self):
        q = ahrs.Quaternion([1., 0., 0., 0.])
        np.testing.assert_allclose(q.rotate([[1., 2.], [3., 4.], [5., 6.]]), np.array([[1., 2.], [3., 4.], [5., 6.]]))
        q = ahrs.Quaternion([1., 0., 1., 0.])
        np.testing.assert_allclose(q.rotate([[1., 2.], [3., 4.], [5., 6.]]), np.array([[5., 6.], [3., 4.], [-1., -2.]]))
        q = ahrs.Quaternion([0., 0., 1., 0.])
        np.testing.assert_allclose(q.rotate([[1., 2.], [3., 4.], [5., 6.]]), np.array([[-1., -2.], [3., 4.], [-5., -6.]]))

    def test_wrong_input_rotate(self):
        q = ahrs.Quaternion([1., 0., 0., 0.])
        self.assertRaises(ValueError, q.rotate, [1., 2.])
        self.assertRaises(ValueError, q.rotate, [1., 2., 3., 4.])
        self.assertRaises(ValueError, q.rotate, [[1., 2.], [3., 4.]])

class TestQuaternionArray(unittest.TestCase):
    def setUp(self) -> None:
        self.Q0 = ahrs.QuaternionArray()
        self.Q1 = ahrs.QuaternionArray(np.identity(4))
        self.decimal_precision = 15

    def test_identity_quaternion(self):
        self.assertEqual(self.Q0.w, [1.0])
        self.assertEqual(self.Q0.x, [0.0])
        self.assertEqual(self.Q0.y, [0.0])
        self.assertEqual(self.Q0.z, [0.0])
        np.testing.assert_equal(self.Q0.v, np.atleast_2d(np.zeros(3)))

    def test_4_by_4_array(self):
        np.testing.assert_equal(self.Q1.w, [1.0, 0.0, 0.0, 0.0])
        np.testing.assert_equal(self.Q1.x, [0.0, 1.0, 0.0, 0.0])
        np.testing.assert_equal(self.Q1.y, [0.0, 0.0, 1.0, 0.0])
        np.testing.assert_equal(self.Q1.z, [0.0, 0.0, 0.0, 1.0])
        np.testing.assert_equal(self.Q1.v, np.eye(4, 3, k=-1))

    def test_quaternion_average(self):
        # Create 10 continuous quaternions representing rotations around the
        # Z-axis between 0 and 90 degrees and average them. The mean quaternion
        # must be equal to the quaternion representing a rotation of 45 degrees
        # around the Z-axis.
        Q = ahrs.QuaternionArray([ahrs.Quaternion(rpy=np.array([0.0, 0.0, x])*ahrs.DEG2RAD) for x in range(0, 100, 10)])
        np.testing.assert_almost_equal(Q.average(), ahrs.Quaternion(rpy=np.array([0.0, 0.0, 45.0])*ahrs.DEG2RAD), decimal=self.decimal_precision)

    def test_wrong_input_array(self):
        self.assertRaises(TypeError, ahrs.QuaternionArray, True)
        self.assertRaises(TypeError, ahrs.QuaternionArray, 3.0)
        self.assertRaises(TypeError, ahrs.QuaternionArray, "[[1.0, 2.0, 3.0, 4.0]]")
        self.assertRaises(ValueError, ahrs.QuaternionArray, np.random.random(4))
        self.assertRaises(ValueError, ahrs.QuaternionArray, [[1., 2., 3., 4.], [1., 2., 3.]])

    def test_is_pure(self):
        Q = ahrs.QuaternionArray([[0., 1., 2., 3.], [1., 2., 3., 4.], [0., 0., 0., 1.]])
        self.assertListEqual(Q.is_pure().tolist(), [True, False, True])

    def test_is_real(self):
        Q = ahrs.QuaternionArray([[1., 0., 0., 0.], [1., 2., 3., 4.], [-1., 0., 0., 0.]])
        self.assertListEqual(Q.is_real().tolist(), [True, False, True])

    def test_is_versor(self):
        Q = ahrs.QuaternionArray([[1., 0., 0., 0.], [1., 2., 3., 4.], [0., 0., 0., 1.]], versors=False)
        self.assertListEqual(Q.is_versor().tolist(), [True, False, True])
        Q = ahrs.QuaternionArray([[1., 0., 0., 0.], [1., 2., 3., 4.], [0., 0., 0., 1.]])
        self.assertListEqual(Q.is_versor().tolist(), [True, True, True])

    def test_is_identity(self):
        Q = ahrs.QuaternionArray([[1., 0., 0., 0.], [1., 2., 3., 4.], [0., 0., 0., 1.]])
        self.assertListEqual(Q.is_identity().tolist(), [True, False, False])
        Q = ahrs.QuaternionArray([[1., 0., 0., 0.], [-1., 0., 0., 0.], [2., 0., 0., 0.]])
        self.assertListEqual(Q.is_identity().tolist(), [True, False, True])

    def test_conjugate(self):
        Q = ahrs.QuaternionArray([[1., 0., 0., 0.], [1., 2., 3., 4.], [0., 0., 0., 1.]])
        Q_conj = ahrs.QuaternionArray([[1., 0., 0., 0.], [1., -2., -3., -4.], [0., 0., 0., -1.]])
        np.testing.assert_almost_equal(Q.conjugate(), Q_conj, decimal=self.decimal_precision)

    def test_conj(self):
        Q = ahrs.QuaternionArray([[1., 0., 0., 0.], [1., 2., 3., 4.], [0., 0., 0., 1.]])
        Q_conj = ahrs.QuaternionArray([[1., 0., 0., 0.], [1., -2., -3., -4.], [0., 0., 0., -1.]])
        np.testing.assert_almost_equal(Q.conj(), Q_conj, decimal=self.decimal_precision)

    def test_to_DCM(self):
        R = self.Q1.to_DCM()
        np.testing.assert_equal(R[0], np.identity(3))
        np.testing.assert_equal(R[1], np.diag([1., -1., -1.]))
        np.testing.assert_equal(R[2], np.diag([-1., 1., -1.]))
        np.testing.assert_equal(R[3], np.diag([-1., -1., 1.]))

    def test_slerp_nan(self):
        Q = ahrs.QuaternionArray([[1., 0., 0., 0.], [1., 0., 0., 0.], [0., 0., 0., 1.]])
        Q[1] = np.nan
        Q_slerp = Q.slerp_nan(inplace=False)
        expected_interpolation = np.array([[1, 0, 0, 0],
                                           [SQ22, 0, 0, SQ22],
                                           [0, 0, 0, 1]])
        np.testing.assert_almost_equal(Q_slerp, expected_interpolation)

class TestSLERP(unittest.TestCase):
    def setUp(self) -> None:
        self.q1 = ahrs.Quaternion([1.0, 0.0, 0.0, 0.0])
        self.q2 = ahrs.Quaternion([0.0, 0.0, 0.0, 1.0])
        self.q3 = ahrs.Quaternion([1.0, 0.0, 0.0, 1.0])
        self.q4 = ahrs.Quaternion([1.0, 0.0, 1.0, 0.0])
        self.q5 = ahrs.Quaternion([1.0, 0.0, 0.0, 0.95])

    def test_slerp_interpolations(self):
        q_slerp = ahrs.common.quaternion.slerp(self.q1, self.q2, [0, 0.5, 1])
        expected_interpolation = np.array([[1, 0, 0, 0],
                                           [SQ22, 0, 0, SQ22],
                                           [0, 0, 0, 1]])
        np.testing.assert_almost_equal(q_slerp, expected_interpolation)
        q_slerp = ahrs.common.quaternion.slerp(self.q3, -self.q4, [0, 0.5, 1])
        expected_interpolation = np.array([[SQ22, 0, 0, SQ22],
                                           [1/np.sqrt(1.5), 0, 0.5/np.sqrt(1.5), 0.5/np.sqrt(1.5)],
                                           [SQ22, 0, SQ22, 0]])
        np.testing.assert_almost_equal(q_slerp, expected_interpolation)

    def test_lerp_interpolations(self):
        q_slerp = ahrs.common.quaternion.slerp(self.q3, self.q5, [0, 0.5, 1])
        midq = np.array([(self.q3[0]+self.q5[0])/2, 0., 0., (self.q3[3]+self.q5[3])/2])
        midq /= np.linalg.norm(midq)
        expected_interpolation = np.vstack((self.q3, midq, self.q5))
        np.testing.assert_almost_equal(q_slerp, expected_interpolation)

class TestRandomAttitudes(unittest.TestCase):
    def setUp(self) -> None:
        self.num_attitudes = 10

    def test_one_random_attitude(self):
        q = ahrs.common.quaternion.random_attitudes(1)
        self.assertEqual(q.shape, (4,))
        rotmat = ahrs.common.quaternion.random_attitudes(1, representation="rotmat")
        self.assertEqual(rotmat.shape, (3, 3))
        self.assertTrue(np.allclose(np.linalg.det(rotmat), 1))
        self.assertTrue(np.allclose(rotmat@rotmat.T, np.identity(3)))

    def test_many_random_attitudes(self):
        q = ahrs.common.quaternion.random_attitudes(self.num_attitudes)
        self.assertEqual(q.shape, (self.num_attitudes, 4))
        self.assertTrue(np.all(np.abs(np.linalg.norm(q, axis=1)-1) < 1e-15))
        rotmats = ahrs.common.quaternion.random_attitudes(self.num_attitudes, representation="rotmat")
        self.assertEqual(rotmats.shape, (self.num_attitudes, 3, 3))
        self.assertTrue(np.allclose(np.linalg.det(rotmats), np.ones(self.num_attitudes)))
        self.assertTrue(np.allclose([x@x.T for x in rotmats], np.identity(3)))

    def test_wrong_input(self):
        self.assertRaises(ValueError, ahrs.common.quaternion.random_attitudes, 0)
        self.assertRaises(ValueError, ahrs.common.quaternion.random_attitudes, -1)
        self.assertRaises(TypeError, ahrs.common.quaternion.random_attitudes, [10])
        self.assertRaises(TypeError, ahrs.common.quaternion.random_attitudes, (10,))
        self.assertRaises(TypeError, ahrs.common.quaternion.random_attitudes, np.array([10]))
        self.assertRaises(TypeError, ahrs.common.quaternion.random_attitudes, 1.0)
        self.assertRaises(TypeError, ahrs.common.quaternion.random_attitudes, "10")
        self.assertRaises(ValueError, ahrs.common.quaternion.random_attitudes, 10, representation="euler")
        self.assertRaises(ValueError, ahrs.common.quaternion.random_attitudes, 10, representation="dcm")
        self.assertRaises(TypeError, ahrs.common.quaternion.random_attitudes, 10, representation=0)
        self.assertRaises(TypeError, ahrs.common.quaternion.random_attitudes, 10, representation=1.0)
        self.assertRaises(TypeError, ahrs.common.quaternion.random_attitudes, 10, representation=["quaternion"])
        self.assertRaises(TypeError, ahrs.common.quaternion.random_attitudes, 10, representation=("quaternion",))

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
import unittest
import ahrs

class TestWELMEC(unittest.TestCase):
    def test_correct_values(self):
        self.assertAlmostEqual(ahrs.utils.welmec_gravity(52.3, 80.0), 9.812484, places=6)     # Braunschweig, Germany
        self.assertAlmostEqual(ahrs.utils.welmec_gravity(60.0, 250.0), 9.818399, places=6)    # Uppsala, Sweden

class TestInternationalGravity(unittest.TestCase):
    def test_correct_values(self):
        self.assertEqual(ahrs.utils.international_gravity(10.0), 9.781924844039924)
        self.assertEqual(ahrs.utils.international_gravity(10.0, epoch='1930'), 9.7820428934191)
        self.assertEqual(ahrs.utils.international_gravity(10.0, epoch='1984'), 9.781882458192761)

class TestWGS84(unittest.TestCase):
    def setUp(self) -> None:
        self.wgs = ahrs.utils.WGS()

    def test_normal_gravity(self):
        self.assertEqual(self.wgs.normal_gravity(50.0), 9.810702135603085)
        self.assertEqual(self.wgs.normal_gravity(50.0, 100.0), 9.810393625316983)
        self.assertAlmostEqual(self.wgs.normal_gravity(0.0), 9.7803253359, places=10)
        self.assertAlmostEqual(self.wgs.normal_gravity(90.0), 9.8321849379, places=10)

    def test_values(self):
        self.assertAlmostEqual(self.wgs.a, 6_378_137.0, 1)
        self.assertAlmostEqual(self.wgs.f, 1/298.257223563, 7)
        self.assertAlmostEqual(self.wgs.gm, 3.986004418e14, 7)
        self.assertAlmostEqual(self.wgs.w, 7.292115e-5, 7)
        self.assertAlmostEqual(self.wgs.b, 6_356_752.3142, 4)
        self.assertAlmostEqual(self.wgs.is_geodetic, True)
        self.assertAlmostEqual(self.wgs.first_eccentricity_squared, 8.1819190842622e-2**2, 7)
        self.assertAlmostEqual(self.wgs.second_eccentricity_squared, 8.2094437949696e-2**2, 7)
        self.assertAlmostEqual(self.wgs.linear_eccentricity, 5.2185400842339e5, 7)
        self.assertAlmostEqual(self.wgs.aspect_ratio, 9.966471893352525e-1, 10)
        self.assertAlmostEqual(self.wgs.curvature_polar_radius, 6399593.6258, 4)
        self.assertAlmostEqual(self.wgs.arithmetic_mean_radius, 6371008.7714, 4)
        self.assertAlmostEqual(self.wgs.authalic_sphere_radius, 6371007.181, 3)
        self.assertAlmostEqual(self.wgs.equivolumetric_sphere_radius, 6371000.79, 2)
        self.assertAlmostEqual(self.wgs.normal_gravity_constant, 3.44978650684084e-3, 10)
        self.assertAlmostEqual(self.wgs.dynamical_form_factor, 1.082629821313e-3, 10)
        self.assertAlmostEqual(self.wgs.second_degree_zonal_harmonic, -4.84166774985e-4, 10)
        self.assertAlmostEqual(self.wgs.normal_gravity_potential, 6.26368517146e7, 4)
        self.assertAlmostEqual(self.wgs.equatorial_normal_gravity, 9.7803253359, 10)
        self.assertAlmostEqual(self.wgs.polar_normal_gravity, 9.8321849379, 10)
        self.assertAlmostEqual(self.wgs.mean_normal_gravity, 9.7976432223, 10)
        self.assertAlmostEqual(self.wgs.mass, 5.9721864e24, delta=1e17)
        self.assertAlmostEqual(self.wgs.atmosphere_gravitational_constant, 3.43592e8, delta=1e3)
        self.assertAlmostEqual(self.wgs.gravitational_constant_without_atmosphere, 3.986000982e14, delta=1e4)
        self.assertAlmostEqual(self.wgs.dynamic_inertial_moment_about_Z, 8.03430094201443e37, delta=1e30)   # FAILS with report's reference
        self.assertAlmostEqual(self.wgs.dynamic_inertial_moment_about_X, 8.00792178e37, delta=1e29)
        self.assertAlmostEqual(self.wgs.dynamic_inertial_moment_about_Y, 8.0080748e37, delta=1e30)
        self.assertAlmostEqual(self.wgs.geometric_inertial_moment_about_Z, 8.07302937e37, delta=1e29)
        self.assertAlmostEqual(self.wgs.geometric_inertial_moment, 8.04672663e37, delta=1e29)
        self.assertAlmostEqual(self.wgs.geometric_dynamic_ellipticity, 3.2581004e-3, 9)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Constants
=========

Common constant values used in AHRS and Geodesy. The constants are defined in
SI Units (second, metre, kilogram) unless otherwise noted, or when constants
are unitless.

**Basic Trigonometry**

===========  ==================  =====================
Name         Description         Value
===========  ==================  =====================
``M_PI``     Pi                  ``3.141592653589793``
``DEG2RAD``  Degrees to Radians  ``0.017453292519943``
``RAD2DEG``  Radians to Degrees  ``57.29577951308232``
===========  ==================  =====================

**Geodesy**

The following constants are set as defined in the latest report of the World
Geodetic System 1984 :cite:p:`wgs84-2014` and can be compared against
:cite:p:`chambat2001`. The CODATA constants :cite:p:`mohr2016` and
:cite:p:`tiesinga2021` are differentiated with an ending indicating their epoch
as ``CODATA2014`` and ``CODATA2018`` :cite:p:`codata2018` respectively.

====================================  ==============================================  =============
Name                                  Description                                     Value
====================================  ==============================================  =============
``DYNAMIC_ELLIPTICITY``               Dynamic Ellipticity                             ``3.2737949e-3``
``EARTH_ATMOSPHERE_MASS``             Mass of Atmosphere including water vapor        ``5.148e18``
``EARTH_AUTHALIC_RADIUS``             Earth's Radius of equal area sphere             ``6371007.1810``
``EARTH_AXIS_RATIO``                  Earth's Axis ratio                              ``9.96647189335e-1``
``EARTH_C20_DYN``                     Earth's Dynamic 2nd Degree Zonal Harmonic       ``-4.84165143790815e-4``
``EARTH_C22_DYN``                     Earth's Dynamic 2nd Degree Sectorial Harmonic   ``2.43938357328313e-6``
``EARTH_C20_GEO``                     Earth's Geographic 2nd Degree Zonal Harmonic    ``-4.84166774985e-4``
``EARTH_EQUATOR_RADIUS``              Earth's Semi-major axis (Equatorial Radius)     ``6378137.0``
``EARTH_EQUIVOLUMETRIC_RADIUS``       Earth's Radius of equal volume sphere           ``6371000.79``
``EARTH_FIRST_ECCENTRICITY``          Earth's First Eccentricity                      ``8.1819190842622e-2``
``EARTH_FIRST_ECCENTRICITY_2``        Earth's First Eccentricity Squared              ``6.6943799901414e-3``
``EARTH_FLATTENING_INV``              Earth's Flattening Factor                       ``298.257223563``
``EARTH_FLATTENING``                  Earth's Flattening Factor (reduced)             ``1/298.257223563``
``EARTH_GM``                          Earth's Gravitational Constant (GM)             ``3.986004418e14``
``EARTH_GM_1``                        Earth's GM without Atmosphere                   ``3.986000982e14``
``EARTH_GM_2``                        Earth Atmosphere's Gravitational Constant       ``3.4359e8``
``EARTH_GM_GPSNAV``                   Earth's GM for GPS Navigation                   ``3.9860050e14``
``EARTH_J2``                          Earth's Dynamic Form Factor                     ``1.08263e-3``
``EARTH_LINEAR_ECCENTRICITY``         Earth's Linear Eccentricity                     ``5.2185400842339e5``
``EARTH_MEAN_RADIUS``                 Earth's Arithmetic Mean radius                  ``6371200.0``
``EARTH_MEAN_AXIAL_RADIUS``           Earth's Mean Radius of the three semi-axes      ``6371008.7714``
``EARTH_MASS``                        Earth's Mass (Atmosphere inclulded)             ``5.9721864e24``
``EARTH_POLAR_CURVATURE_RADIUS``      Earth's Polar Radius of Curvature               ``6_399_593.6258``
``EARTH_POLAR_RADIUS``                Earth's Semi-minor axis (Polar Radius)          ``6356752.3142``
``EARTH_ROTATION``                    Earth's Rotation rate                           ``7.292115e-5``
``EARTH_SECOND_ECCENTRICITY``         Earth's Second Eccentricity                     ``8.2094437949696e-2``
``EARTH_SECOND_ECCENTRICITY_2``       Earth's Second Eccentricity Squared             ``6.739496742276486e-3``
``EARTH_SIDEREAL_DAY``                Earth's duration of sidereal day                ``86164.09053083288``
``EQUATORIAL_NORMAL_GRAVITY``         Earth's Normal Gravity at the Equator           ``9.7803253359``
``LIGHT_SPEED``                       Velocity of light in vacuum                     ``2.99792458e8``
``MEAN_NORMAL_GRAVITY``               Earth's Mean Normal Gravity                     ``9.7976432223``
``NORMAL_GRAVITY_FORMULA``            Constant for Normal Gravity Formula             ``3.449786506841e-3``
``NORMAL_GRAVITY_POTENTIAL``          Earth's Normal Gravity Potential                ``6.26368517146``
``POLAR_NORMAL_GRAVITY``              Earth's Normal Gravity at the Pole              ``9.8321849379``
``SOMIGLIANA_GRAVITY``                Constant for Somigliana's Formula               ``1.931852652458e-3``
``UNIVERSAL_GRAVITATION_CODATA2014``  Universal Gravitation defined in CODATA2014     ``6.67408e-11``
``UNIVERSAL_GRAVITATION_CODATA2018``  Universal Gravitation defined in CODATA2018     ``6.67430e-11``
``UNIVERSAL_GRAVITATION_WGS84``       Universal Gravitation defined in WGS84          ``6.67428e-11``
====================================  ==============================================  =============

The values above are set as defined in :cite:p:`wgs84-2014`, although most of
these values can be also obtained with the class ``WGS`` of this package, which
builds the World's Geodetic System independently.

The elemental defining parameters (equatorial radius, flattening, gravitational
constant and rotational velocity) are set, by default, to that of Earth's, and
all other parameters are derived from these. Just to compare:

.. code:: python

    >>> import ahrs
    >>> ahrs.EARTH_EQUATOR_RADIUS
    6378137.0
    >>> ahrs.EARTH_POLAR_RADIUS
    6356752.3142
    >>> ahrs.EARTH_FIRST_ECCENTRICITY_2
    0.0066943799901414
    >>> wgs = ahrs.utils.WGS()  # Default model is Earth's
    >>> wgs.a
    6378137.0
    >>> wgs.b
    6356752.314245179
    >>> wgs.first_eccentricity_squared
    0.0066943799901413165

These values are, so far, determined for the Earth. However, other celestial
bodies have been measured and their values are, as defined in
:cite:p:`archinal2018`, :cite:p:`park2019` and :cite:p:`williams2024`, set as
follows:

==========================  ==========================
Name                        Value
==========================  ==========================
``MOON_EQUATOR_RADIUS``     ``1738100.0``
``MOON_POLAR_RADIUS``       ``1736000.0``
``MOON_MASS``               ``7.346e22``
``MOON_ROTATION``           ``1.109027709148159e-7``
``MOON_J2``                 ``2.027e-4``
``MERCURY_EQUATOR_RADIUS``  ``2440530.0``
``MERCURY_POLAR_RADIUS``    ``2438260.0``
``MERCURY_ROTATION``        ``1.2399326882596827e-6``
``MERCURY_MASS``            ``3.30114e23``
``MERCURY_J2``              ``5.03e-5``
``VENUS_EQUATOR_RADIUS``    ``6051800.0``
``VENUS_POLAR_RADIUS``      ``6051800.0``
``VENUS_ROTATION``          ``-2.9923691869737844e-7``
``VENUS_MASS``              ``4.86747e24``
``VENUS_J2``                ``4.458e-6``
``MARS_EQUATOR_RADIUS``     ``3396190.0``
``MARS_POLAR_RADIUS``       ``3376200.0``
``MARS_ROTATION``           ``7.088235959185674e-5``
``MARS_MASS``               ``6.41712e23``
``MARS_J2``                 ``1.96045e-3``
``JUPITER_EQUATOR_RADIUS``  ``71492000.0``
``JUPITER_POLAR_RADIUS``    ``66854000.0``
``JUPITER_ROTATION``        ``1.758518138029551e-4``
``JUPITER_MASS``            ``1.898187e27``
``JUPITER_J2``              ``1.4736e-2``
``SATURN_EQUATOR_RADIUS``   ``60268000.0``
``SATURN_POLAR_RADIUS``     ``54364000.0``
``SATURN_ROTATION``         ``1.637884057802486e-4``
``SATURN_MASS``             ``5.683174e26``
``SATURN_J2``               ``1.6298e-2``
``URANUS_EQUATOR_RADIUS``   ``25559000.0``
``URANUS_POLAR_RADIUS``     ``24973000.0``
``URANUS_ROTATION``         ``-1.012376653716682e-4``
``URANUS_MASS``             ``8.68127e25``
``URANUS_J2``               ``3.343430e-3``
``NEPTUNE_EQUATOR_RADIUS``  ``24764000.0``
``NEPTUNE_POLAR_RADIUS``    ``24341000.0``
``NEPTUNE_ROTATION``        ``1.083382527619075e-4``
``NEPTUNE_MASS``            ``1.024126e26``
``NEPTUNE_J2``              ``3.411e-3``
``PLUTO_EQUATOR_RADIUS``    ``1188300.0``
``PLUTO_POLAR_RADIUS``      ``1188300.0``
``PLUTO_ROTATION``          ``-1.138559183467410e-05``
``PLUTO_MASS``              ``1.303e22``
==========================  ==========================

"""

import math

# TRIGONOMETRY
M_PI = math.pi
DEG2RAD = M_PI/180
RAD2DEG = 180/M_PI

##### Geodetic constants as defined in WORLD GEODETIC SYSTEM 1984 (rev. 2014)
# Defining parameters
EARTH_EQUATOR_RADIUS = 6_378_137.0              # Semi-major axis of Earth (Equatorial Radius) [m]
EARTH_FLATTENING_INV = 298.257223563            # Flattening Factor of the Earth
EARTH_GM = 3.986004418e14                       # Earth's Gravitational Constant (Atmosphere included) [m^3/s^2]
EARTH_ROTATION = 7.292115e-5                    # Earth's Rotation rate [rad/s]
# Fundamental constants
LIGHT_SPEED = 2.99792458e8                      # Velocity of light in vacuum [m/s]
EARTH_ATMOSPHERE_MASS = 5.148e18                # Total mean mass of the Atmosphere (with water vapor) [kg]
DYNAMIC_ELLIPTICITY = 3.2737949e-3              # Dynamic Ellipticity (H)
# Universal Constant of Gravitation [m^3/(kg*s^2)]
UNIVERSAL_GRAVITATION_WGS84 = 6.67428e-11       # As defined in latest report of WGS 84
UNIVERSAL_GRAVITATION_CODATA2018 = 6.67430e-11  # As recommended by CODATA 2018 in latest report of NIST 2019
UNIVERSAL_GRAVITATION_CODATA2014 = 6.67408e-11  # As recommended by CODATA 2014 and referenced by NASA's Jet Propulsion Laboratory
EARTH_GM_GPSNAV = 3.9860050e14                  # Earth's Gravitational Constant for GPS Navigation Message [m^3/s^2]
# Derived geometric constants
EARTH_FLATTENING = 1/EARTH_FLATTENING_INV       # Earth's Flattening (reduced)
EARTH_POLAR_RADIUS = 6_356_752.3142             # Semi-minor axis of Earth (Polar Radius) [m]
EARTH_FIRST_ECCENTRICITY = 8.1819190842622e-2
EARTH_FIRST_ECCENTRICITY_2 = EARTH_FIRST_ECCENTRICITY**2
EARTH_SECOND_ECCENTRICITY = 8.2094437949696e-2
EARTH_SECOND_ECCENTRICITY_2 = EARTH_SECOND_ECCENTRICITY**2
EARTH_LINEAR_ECCENTRICITY = 5.2185400842339e5
EARTH_POLAR_CURVATURE_RADIUS = 6_399_593.6258   # Polar radius of Curvature [m]
EARTH_AXIS_RATIO = 9.96647189335e-1             # Axis ratio: EARTH_POLAR_RADIUS / EARTH_EQUATOR_RADIUS
EARTH_MEAN_RADIUS = 6_371_200.0                 # Earth's Arithmetic Mean radius [m] ((2*EQUATOR_RADIUS + POLAR_RADIUS) / 3)
EARTH_MEAN_AXIAL_RADIUS = 6_371_008.7714        # Mean Radius of the Three Semi-axes [m]
EARTH_AUTHALIC_RADIUS = 6_371_007.1810          # Radius of equal area sphere [m]
EARTH_EQUIVOLUMETRIC_RADIUS = 6_371_000.79      # Radius of equal volume sphere [m] ((EQUATOR_RADIUS^2 * POLAR_RADIUS)^(1/3))
EARTH_C20_DYN = -4.84165143790815e-4            # Earth's Dynamic Second Degree Zonal Harmonic (C_2,0 dyn)
EARTH_C22_DYN = 2.43938357328313e-6             # Earth's Dynamic Second Degree Sectorial Harmonic (C_2,2 dyn)
EARTH_C20_GEO = -4.84166774985e-4               # Earth's Geographic Second Degree Zonal Harmonic
EARTH_J2 = 1.08263e-3                           # Earth's Dynamic Form Factor
# Derived physical constants
NORMAL_GRAVITY_POTENTIAL = 6.26368517146        # Normal Gravity Potential on the Ellipsoid [m^2/s^2]
EQUATORIAL_NORMAL_GRAVITY = 9.7803253359        # Normal Gravity at the Equator (on the ellipsoid) [m/s^2]
POLAR_NORMAL_GRAVITY = 9.8321849379             # Normal Gravity at the Pole (on the ellipsoid) [m/s^2]
MEAN_NORMAL_GRAVITY = 9.7976432223              # Mean Normal Gravity [m/s^2]
SOMIGLIANA_GRAVITY = 1.931852652458e-3          # Somigliana's Formula Normal Gravity constant
NORMAL_GRAVITY_FORMULA = 3.449786506841e-3      # Normal Gravity Formula constant (EARTH_ROTATION^2 * EQUATOR_RADIUS^2 * POLAR_RADIUS / EARTH_GM)
EARTH_MASS = 5.9721864e24                       # Earth's Mass (Atmosphere inclulded) [kg]
EARTH_GM_1 = 3.986000982e14                     # Geocentric Gravitational Constant (Atmosphere excluded) [m^3/s^2]
EARTH_GM_2 = 3.4359e8                           # Gravitational Constant of the Earth’s Atmosphere [m^3/s^2]

EARTH_SIDEREAL_DAY = 86164.09053083288          # Earth's duration of sidereal day [s]
##### Planetary Characteristics (without Earth)
MOON_EQUATOR_RADIUS = 1_738_100.0
MOON_POLAR_RADIUS = 1_736_000.0
MOON_MASS = 7.346e22
MOON_GM = MOON_MASS*UNIVERSAL_GRAVITATION_CODATA2018
MOON_ROTATION = 1.109027709148159e-7            # Inferred from [7]
MOON_J2 = 2.027e-4                              # As defined in [7]
MERCURY_EQUATOR_RADIUS = 2_440_530.0            # As defined in [6]
MERCURY_POLAR_RADIUS = 2_438_260.0
MERCURY_ROTATION = 1.2399326882596827e-6        # Inferred from [7]
MERCURY_MASS = 3.30114e23                       # As defined in [6]
MERCURY_GM = MERCURY_MASS*UNIVERSAL_GRAVITATION_CODATA2018
MERCURY_J2 = 5.03e-5                            # As defined in [7]
VENUS_EQUATOR_RADIUS = 6_051_800.0              # As defined in [6]
VENUS_POLAR_RADIUS = 6_051_800.0
VENUS_ROTATION = -2.9923691869737844e-7         # Inferred from [7]
VENUS_MASS = 4.86747e24
VENUS_GM = VENUS_MASS*UNIVERSAL_GRAVITATION_CODATA2018
VENUS_J2 = 4.458e-6                             # As defined in [7]
MARS_EQUATOR_RADIUS = 3_396_190.0               # As defined in [6]
MARS_POLAR_RADIUS = 3_376_200.0
MARS_ROTATION = 7.088235959185674e-5
MARS_MASS = 6.41712e23
MARS_GM = MARS_MASS*UNIVERSAL_GRAVITATION_CODATA2018
MARS_J2 = 1.96045e-3                            # As defined in [7]
JUPITER_EQUATOR_RADIUS = 71_492_000.0           # As defined in [6]
JUPITER_POLAR_RADIUS = 66_854_000.0
JUPITER_ROTATION = 1.758518138029551e-4         # Inferred from [7]
JUPITER_MASS = 1.898187e27
JUPITER_GM = JUPITER_MASS*UNIVERSAL_GRAVITATION_CODATA2018
JUPITER_J2 = 1.4736e-2                          # As defined in [7]
SATURN_EQUATOR_RADIUS = 60_268_000.0            # As defined in [6]
SATURN_POLAR_RADIUS = 54_364_000.0
SATURN_ROTATION = 1.637884057802486e-4          # Inferred from [7]
SATURN_MASS = 5.683174e26
SATURN_GM = SATURN_MASS*UNIVERSAL_GRAVITATION_CODATA2018
SATURN_J2 = 1.6298e-2                           # As defined in [7]
URANUS_EQUATOR_RADIUS = 25_559_000.0            # As defined in [6]
URANUS_POLAR_RADIUS = 24_973_000.0
URANUS_ROTATION = -1.012376653716682e-4         # Inferred from [7]
URANUS_MASS = 8.68127e25
URANUS_GM = URANUS_MASS*UNIVERSAL_GRAVITATION_CODATA2018
URANUS_J2 = 3.343430e-3                         # As defined in [7]
NEPTUNE_EQUATOR_RADIUS = 24_764_000.0           # As defined in [6]
NEPTUNE_POLAR_RADIUS = 24_341_000.0
NEPTUNE_ROTATION = 1.083382527619075e-4         # Inferred from [7]
NEPTUNE_MASS = 1.024126e26
NEPTUNE_GM = NEPTUNE_MASS*UNIVERSAL_GRAVITATION_CODATA2018
NEPTUNE_J2 = 3.411e-3                           # As defined in [7]
PLUTO_EQUATOR_RADIUS = 1_188_300.0              # As defined in [6]
PLUTO_POLAR_RADIUS = 1_188_300.0
PLUTO_ROTATION = -1.138559183467410e-05         # Inferred from [7]
PLUTO_MASS = 1.303e22
PLUTO_GM = PLUTO_MASS*UNIVERSAL_GRAVITATION_CODATA2018

##### Local information
# Munich, Germany location of km 0 (Mariensäule)
MUNICH_LATITUDE = 48.137230
MUNICH_LONGITUDE = 11.575508
# Munich, Germany orthometric height (above sea level) in meters
MUNICH_HEIGHT = 521.0 # (https://whatismyelevation.com/map?lat=48.13723&lng=11.57551)
MUNICH_ELEVATION = MUNICH_HEIGHT
MUNICH_ALTITUDE = MUNICH_HEIGHT



# -*- coding: utf-8 -*-
"""
AHRS: Versioning
================

"""

MAJOR = 0
MINOR = 3
PATCH = 1
RELEASE = "0"

def get_version(short: bool = False) -> str:
    if short or RELEASE=="0":
        return f"{MAJOR}.{MINOR}.{PATCH}"
    return f"{MAJOR}.{MINOR}.{PATCH}-{RELEASE}"



#!/usr/bin/env python3
import unittest
import numpy as np
import ahrs

class TestEuclidean(unittest.TestCase):
    def test_correct_values(self):
        self.assertEqual(ahrs.utils.euclidean(np.zeros(3), np.zeros(3)), 0.0)
        self.assertEqual(ahrs.utils.euclidean(np.zeros(3), np.ones(3)), np.sqrt(3))
        self.assertEqual(ahrs.utils.euclidean(np.ones(3), -np.ones(3)), 2.0*np.sqrt(3))
        self.assertEqual(ahrs.utils.euclidean(np.array([1, 2, 3]), np.array([4, 5, 6])), 5.196152422706632)
        self.assertGreaterEqual(ahrs.utils.euclidean(np.random.random(3)-0.5, np.random.random(3)-0.5), 0.0)

    def test_guard_clauses(self):
        self.assertRaises(ValueError, ahrs.utils.euclidean, np.zeros(3), np.zeros(2))

class TestChordal(unittest.TestCase):
    def setUp(self):
        self.R1 = ahrs.DCM(rpy=np.array([10.0, -20.0, 30.0])*ahrs.DEG2RAD)
        self.R2 = ahrs.DCM(rpy=np.array([-10.0, 20.0, -30.0])*ahrs.DEG2RAD)

    def test_correct_values(self):
        self.assertEqual(ahrs.utils.chordal(np.identity(3), np.identity(3)), 0.0)
        self.assertEqual(ahrs.utils.chordal(np.identity(3), -np.identity(3)), 2.0*np.sqrt(3))
        self.assertGreaterEqual(ahrs.utils.chordal(np.random.random((3, 3))-0.5, np.random.random((3, 3))-0.5), 0.0)
        self.assertEqual(ahrs.utils.chordal(self.R1, self.R2), 1.6916338074634352)
        self.assertEqual(ahrs.utils.chordal(self.R1.tolist(), self.R2.tolist()), 1.6916338074634352)

    def test_guard_clauses(self):
        self.assertRaises(TypeError, ahrs.utils.chordal, np.identity(3), 3.0)
        self.assertRaises(TypeError, ahrs.utils.chordal, 3.0, np.identity(3))
        self.assertRaises(TypeError, ahrs.utils.chordal, "np.identity(3)", np.identity(3))
        self.assertRaises(TypeError, ahrs.utils.chordal, np.identity(3), "np.identity(3)")
        self.assertRaises(ValueError, ahrs.utils.chordal, np.identity(3), np.identity(2))
        self.assertRaises(ValueError, ahrs.utils.chordal, np.tile(np.identity(3), (2, 1, 1)), np.tile(np.identity(3), (3, 1, 1)))

class TestIdentityDeviation(unittest.TestCase):
    def setUp(self):
        self.R1 = ahrs.DCM(rpy=np.array([10.0, -20.0, 30.0])*ahrs.DEG2RAD)
        self.R2 = ahrs.DCM(rpy=np.array([-10.0, 20.0, -30.0])*ahrs.DEG2RAD)

    def test_correct_values(self):
        self.assertEqual(ahrs.utils.identity_deviation(np.identity(3), np.identity(3)), 0.0)
        self.assertEqual(ahrs.utils.identity_deviation(np.identity(3), -np.identity(3)), 2.0*np.sqrt(3))
        self.assertGreaterEqual(ahrs.utils.identity_deviation(np.random.random((3, 3))-0.5, np.random.random((3, 3))-0.5), 0.0)
        self.assertEqual(ahrs.utils.identity_deviation(self.R1, self.R2), 1.6916338074634352)
        self.assertEqual(ahrs.utils.identity_deviation(self.R1.tolist(), self.R2.tolist()), 1.6916338074634352)

    def test_guard_clauses(self):
        self.assertRaises(TypeError, ahrs.utils.identity_deviation, np.identity(3), 3.0)
        self.assertRaises(TypeError, ahrs.utils.identity_deviation, 3.0, np.identity(3))
        self.assertRaises(TypeError, ahrs.utils.identity_deviation, "np.identity(3)", np.identity(3))
        self.assertRaises(TypeError, ahrs.utils.identity_deviation, np.identity(3), "np.identity(3)")
        self.assertRaises(ValueError, ahrs.utils.identity_deviation, np.identity(3), np.identity(2))
        self.assertRaises(ValueError, ahrs.utils.identity_deviation, np.zeros((3, 3)), np.zeros((2, 2)))

class TestAngularDistance(unittest.TestCase):
    def setUp(self):
        self.R1 = ahrs.DCM(rpy=np.array([10.0, -20.0, 30.0])*ahrs.DEG2RAD)
        self.R2 = ahrs.DCM(rpy=np.array([-10.0, 20.0, -30.0])*ahrs.DEG2RAD)

    def test_correct_values(self):
        self.assertEqual(ahrs.utils.angular_distance(np.identity(3), np.identity(3)), 0.0)
        self.assertAlmostEqual(ahrs.utils.angular_distance(self.R1, self.R2), 1.8133239804628964)
        self.assertAlmostEqual(ahrs.utils.angular_distance(self.R1.tolist(), self.R2.tolist()), 1.8133239804628964)

    def test_guard_clauses(self):
        self.assertRaises(TypeError, ahrs.utils.angular_distance, np.identity(3), 3.0)
        self.assertRaises(TypeError, ahrs.utils.angular_distance, 3.0, np.identity(3))
        self.assertRaises(TypeError, ahrs.utils.angular_distance, "np.identity(3)", np.identity(3))
        self.assertRaises(TypeError, ahrs.utils.angular_distance, np.identity(3), "np.identity(3)")
        self.assertRaises(ValueError, ahrs.utils.angular_distance, np.identity(3), np.identity(2))
        self.assertRaises(ValueError, ahrs.utils.angular_distance, np.zeros((3, 3)), np.zeros((2, 2)))

class TestQdist(unittest.TestCase):
    def setUp(self):
        self.q1 = ahrs.Quaternion([0., 1., 0., 1.])
        self.q2 = ahrs.Quaternion([1., 0., 1., 0.])

    def test_correct_values(self):
        self.assertEqual(ahrs.utils.qdist(ahrs.Quaternion(), [1., 0., 0., 0.]), 0.0)
        self.assertEqual(ahrs.utils.qdist(self.q1, [0., 1., 0., 1.]), 0.0)
        self.assertEqual(ahrs.utils.qdist(self.q1, [0., 0.7071, 0., 0.7071]), 0.0)
        self.assertAlmostEqual(ahrs.utils.qdist(self.q1, self.q2), np.sqrt(2))

    def test_guard_clauses(self):
        self.assertRaises(TypeError, ahrs.utils.qdist, ahrs.Quaternion(), 3.0)
        self.assertRaises(TypeError, ahrs.utils.qdist, 3.0, ahrs.Quaternion())
        self.assertRaises(TypeError, ahrs.utils.qdist, "ahrs.Quaternion()", ahrs.Quaternion())
        self.assertRaises(TypeError, ahrs.utils.qdist, ahrs.Quaternion(), "ahrs.Quaternion()")
        self.assertRaises(ValueError, ahrs.utils.qdist, ahrs.Quaternion(), np.identity(2))
        self.assertRaises(ValueError, ahrs.utils.qdist, np.zeros(4), np.zeros(3))

class TestQeip(unittest.TestCase):
    def setUp(self):
        self.q1 = ahrs.Quaternion([0., 1., 0., 1.])
        self.q2 = ahrs.Quaternion([1., 0., 1., 0.])

    def test_correct_values(self):
        self.assertEqual(ahrs.utils.qeip(ahrs.Quaternion(), [1., 0., 0., 0.]), 0.0)
        self.assertEqual(ahrs.utils.qeip(self.q1, [0., 1., 0., 1.]), 0.0)
        self.assertEqual(ahrs.utils.qeip(self.q1, [0., 0.7071, 0., 0.7071]), 0.0)
        self.assertAlmostEqual(ahrs.utils.qeip(self.q1, self.q2), 1.0)

    def test_guard_clauses(self):
        self.assertRaises(TypeError, ahrs.utils.qeip, ahrs.Quaternion(), 3.0)
        self.assertRaises(TypeError, ahrs.utils.qeip, 3.0, ahrs.Quaternion())
        self.assertRaises(TypeError, ahrs.utils.qeip, "ahrs.Quaternion()", ahrs.Quaternion())
        self.assertRaises(TypeError, ahrs.utils.qeip, ahrs.Quaternion(), "ahrs.Quaternion()")
        self.assertRaises(ValueError, ahrs.utils.qeip, ahrs.Quaternion(), np.identity(2))
        self.assertRaises(ValueError, ahrs.utils.qeip, np.zeros(4), np.zeros(3))

class TestQcip(unittest.TestCase):
    def setUp(self):
        self.q1 = ahrs.Quaternion([0., 1., 0., 1.])
        self.q2 = ahrs.Quaternion([1., 0., 1., 0.])

    def test_correct_values(self):
        self.assertEqual(ahrs.utils.qcip(ahrs.Quaternion(), [1., 0., 0., 0.]), 0.0)
        self.assertEqual(ahrs.utils.qcip(self.q1, [0., 1., 0., 1.]), 0.0)
        self.assertEqual(ahrs.utils.qcip(self.q1, [0., 0.7071, 0., 0.7071]), 0.0)
        self.assertAlmostEqual(ahrs.utils.qcip(self.q1, self.q2), np.pi/2)

    def test_guard_clauses(self):
        self.assertRaises(TypeError, ahrs.utils.qcip, ahrs.Quaternion(), 3.0)
        self.assertRaises(TypeError, ahrs.utils.qcip, 3.0, ahrs.Quaternion())
        self.assertRaises(TypeError, ahrs.utils.qcip, "ahrs.Quaternion()", ahrs.Quaternion())
        self.assertRaises(TypeError, ahrs.utils.qcip, ahrs.Quaternion(), "ahrs.Quaternion()")
        self.assertRaises(ValueError, ahrs.utils.qcip, ahrs.Quaternion(), np.identity(2))
        self.assertRaises(ValueError, ahrs.utils.qcip, np.zeros(4), np.zeros(3))

class TestQad(unittest.TestCase):
    def setUp(self):
        self.q1 = ahrs.Quaternion([0., 1., 0., 1.])
        self.q2 = ahrs.Quaternion([1., 0., 1., 0.])

    def test_correct_values(self):
        self.assertEqual(ahrs.utils.qad(ahrs.Quaternion(), [1., 0., 0., 0.]), 0.0)
        self.assertEqual(ahrs.utils.qad(self.q1, [0., 1., 0., 1.]), 0.0)
        self.assertEqual(ahrs.utils.qad(self.q1, [0., 0.7071, 0., 0.7071]), 0.0)
        self.assertAlmostEqual(ahrs.utils.qad(self.q1, self.q2), np.pi)

    def test_guard_clauses(self):
        self.assertRaises(TypeError, ahrs.utils.qad, ahrs.Quaternion(), 3.0)
        self.assertRaises(TypeError, ahrs.utils.qad, 3.0, ahrs.Quaternion())
        self.assertRaises(TypeError, ahrs.utils.qad, "ahrs.Quaternion()", ahrs.Quaternion())
        self.assertRaises(TypeError, ahrs.utils.qad, ahrs.Quaternion(), "ahrs.Quaternion()")
        self.assertRaises(ValueError, ahrs.utils.qad, ahrs.Quaternion(), np.identity(2))
        self.assertRaises(ValueError, ahrs.utils.qad, np.zeros(4), np.zeros(3))

if __name__ == "__main__":
    unittest.main()



"""
`Geodesy <https://en.wikipedia.org/wiki/Geodesy>`_ is the science of measuring
and understanding Earth's shape, size, orientation in space, and gravity field.

A `reference ellipsoid <https://en.wikipedia.org/wiki/Earth_ellipsoid#Reference_ellipsoid>`_
is a simple, smooth shape used to approximate the Earth's true form (or that of
another planet). It is basically a mathematical definition of a planet's
surface defined as an `oblate spheroid <https://en.wikipedia.org/wiki/Spheroid#Oblate_spheroids>`_,
ignoring the bumps and dips caused by mountains, valleys, and other features.

The `geoid <https://en.wikipedia.org/wiki/Geoid>`_ is a version of Earth's
shape, smoothed out and stripped of mountains, valleys, and other surface
features. It represents the average sea level if the oceans were calm, with no
tides, currents, or air pressure changes.

The geoid is uneven because of differences in the planet's gravity caused by
variations in its interior composition and density.

For example, its surface sits higher than the reference ellipsoid in areas
where there's extra mass, causing stronger gravity (a positive gravity anomaly.)

On the other hand, it dips below the reference ellipsoid in places where
there's less mass, leading to weaker gravity (a negative gravity anomaly.)

Such changes are, however, very small and are usually measured in `milligals
<https://en.wikipedia.org/wiki/Gal_(unit)>`_. Thus, they are not considered in
the reference ellipsoid, and neither in this package.

Since reference ellipsoids are easier to work with than the irregular geoid,
they are used for mapping and calculations. They provide the surface on which
coordinates like latitude, longitude, and elevation are defined.

World Models
------------

The two most important geospatial systems used in mapping, navigation, and
Earth science applications are the `World Geodetic System (WGS84)
<https://en.wikipedia.org/wiki/World_Geodetic_System>`_ and the `World Magnetic
Model <https://en.wikipedia.org/wiki/World_Magnetic_Model>`_.

These systems often work together in navigation devices to provide both
positional (WGS84) and directional (WMM) information.

The **WGS84** is a standard reference coordinate system for Earth widely used
for global positioning systems (GPS) and serves as the underlying geodetic
system for latitude, longitude, and altitude coordinates.

WGS84 approximates Earth's shape using a **Reference Ellipsoid** model to
simplify the calculation of other geodetic parameters, such as the Earth's
radius, circumference, and surface area.

The **WMM (World Magnetic Model)** is a mathematical model that represents the
Earth's magnetic field. It is used to calculate magnetic declination (the angle
difference between `true north <https://en.wikipedia.org/wiki/True_north>`_ and
`magnetic north <https://en.wikipedia.org/wiki/North_magnetic_pole>`_.)

It provides magnetic field data for navigation systems, including compasses,
aircraft, and smartphones.

This model is updated every 5 years (e.g., 2020-2025 is the current model) to
account for changes in the Earth's magnetic field.

These three systems (reference ellipsoid, WGS84, and WMM) are implemented in
this package.

The WGS84 model is enough for most applications including the navigation of
aircraft, ships, and cars, as well as for surveying and mapping on Earth.
Therefore, we will use the WGS84 as the default reference ellipsoid throughout
this library for all our implementations.

"""

import numpy as np
from ..common import DEG2RAD
from ..common.constants import EARTH_EQUATOR_RADIUS
from ..common.constants import EARTH_FLATTENING
from ..common.constants import EARTH_GM
from ..common.constants import EARTH_ROTATION
from ..common.constants import UNIVERSAL_GRAVITATION_CODATA2018

# Default values for a basic reference ellipsoid
DEFAULT_MASS = 1.0  # 1 kg
DEFAULT_STANDARD_GRAVITATIONAL_PARAMETER = UNIVERSAL_GRAVITATION_CODATA2018 * DEFAULT_MASS  # GM

class ReferenceEllipsoid:
    """
    This implementation is based on the World Geodetic System (WGS) reference
    ellipsoid :cite:p:`wgs84-2014`, which is the most widely used reference
    ellipsoid for the Global Positioning System (GPS).

    The WGS84 uses four elemental parameters to describe a reference ellipsoid:

    - Semi-major axis, :math:`a`, in meters.
    - Flattening ratio, :math:`f`.
    - Standard Gravitational Parameter, :math:`GM`, in m^3/s^2.
    - Angular velocity, :math:`\\omega`, in rad/s.

    The semi-major axis, :math:`a`, is the radius of the ellipsoid at the
    equator, while the flattening ratio, :math:`f`, is the ratio of the
    difference between the semi-major and semi-minor axes to the semi-major
    axis:

    .. math::

        f = \\frac{a-b}{a}

    where :math:`b` is the semi-minor axis of the ellipsoid. When :math:`f = 0`
    the ellipsoid is a perfect sphere. When :math:`f = 1` the ellipsoid is a
    flat disc.

    The `standard gravitational parameter
    <https://en.wikipedia.org/wiki/Standard_gravitational_parameter>`_,
    :math:`GM`, is the product of the **gravitational constant** and the
    **mass** of the planetary body:

    .. math::

        GM = G \\cdot M

    where :math:`G = 6.67430 \\times 10^{-11} \\frac{\\mathrm{m}^3}{\\mathrm{kg}
    \\mathrm{s}^2}` is the `universal constant of gravitation
    <https://en.wikipedia.org/wiki/Gravitational_constant#Value_and_uncertainty>`_,
    and :math:`M` is the mass of the planetary body, in kg.

    From these four parameters we can derive all other properties of the
    ellipsoid.

    The default values describe a **perfect sphere** with a radius of 1 meter,
    whose gravitational constant is equal the universal constant of gravitation,
    and rotates at 1 rad/s.

    Parameters
    ----------
    a : float, default: 1.0
        Semi-major axis of the ellipsoid, in meters.
    f : float, default: 0.0
        Flattening of the ellipsoid, unitless.
    GM : float, default: 6.6743e-11
        Standard Gravitational Parameter of the ellipsoid, in m^3/s^2.
    w : float, default: 1.0
        Angular velocity of the ellipsoid, in rad/s.
    """
    def __init__(self, a: float = 1.0, f: float = 0.0, GM: float = DEFAULT_STANDARD_GRAVITATIONAL_PARAMETER, w: float = 1.0):
        self.a = a
        self.f = f
        self.gm = GM
        self.w = w
        self.b = self.a*(1-self.f)

    def normal_gravity(self, lat: float, h: float = 0.0) -> float:
        """
        Normal Gravity on (or above) Ellipsoidal Surface

        Estimate the normal gravity on or above the surface of an ellipsoidal
        body using Somigliana's formula (on surface) and a series expansion
        (above surface).

        Somigliana's closed formula as desribed by H. Moritz in
        :cite:p:`moritz1984` is:

        .. math::
            \\mathrm{g} = \\frac{a\\mathrm{g}_e \\cos^2\\phi + b\\mathrm{g}_p\\sin^2\\phi}{\\sqrt{a^2cos^2\\phi + b^2\\sin^2\\phi}}

        For numerical computations, a more convenient form is:

        .. math::
            \\mathrm{g} = \\mathrm{g}_e\\frac{1+k\\sin^2\\phi}{\\sqrt{1-e^2\\sin^2\\phi}}

        with the helper constant :math:`k`:

        .. math::
            k = \\frac{b\\mathrm{g}_p}{a\\mathrm{g}_e}-1

        Parameters
        ----------
        lat: float
            Geographical latitude, in decimal degrees.
        h : float, default: 0.0
            Mean sea level height, in meters.

        Return
        ------
        g : float
            Normal gravity at given point in space, in m/s^2.

        Examples
        --------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.normal_gravity(50.0)
        9.810702135603085
        >>> ref.normal_gravity(50.0, 100.0)
        9.810393625316983

        """
        ge = self.equatorial_normal_gravity
        gp = self.polar_normal_gravity
        if ge is None or gp is None:
            raise ValueError("No valid normal gravity values.")
        lat *= DEG2RAD
        e2 = self.first_eccentricity_squared
        k = (self.b*gp)/(self.a*ge)-1
        sin2 = np.sin(lat)**2
        g = ge*(1+k*sin2)/np.sqrt(1-e2*sin2)                        # Gravity on Ellipsoid Surface (eq. 4-1)
        if h == 0.0:
            return g
        # Normal gravity above surface
        m = self.w**2*self.a**2*self.b/self.gm                      # Gravity constant (eq. B-20)
        g *= 1-2*h*(1+self.f+m-2*self.f*sin2)/self.a + 3.0*h**2/self.a**2   # Gravity Above Ellipsoid (eq. 4-3)
        return g

    def vertical_curvature_radius(self, lat: float) -> float:
        """
        Radius of the curvature in the prime vertical, estimated at a given
        latitude, :math:`\\phi`, as:

        .. math::
            R_N = \\frac{a}{\\sqrt{1-e^2\\sin^2\\phi}}

        Parameters
        ----------
        lat : float
            Geographical latitude, in decimal degrees.
        """
        e = np.sqrt(self.first_eccentricity_squared)
        return self.a/np.sqrt(1-e**2*np.sin(lat)**2)

    def meridian_curvature_radius(self, lat: float) -> float:
        """
        Radius of the curvature in the prime meridian, estimated at a given
        latitude, :math:`\\phi`, as per :cite:p:`jekeli2016`:

        .. math::
            R_M = \\frac{a(1-e^2)}{(1-e^2\\sin^2\\phi)^{\\frac{3}{2}}}

        Parameters
        ----------
        lat : float
            Geographical latitude, in decimal degrees.
        """
        e = np.sqrt(self.first_eccentricity_squared)
        return self.a*(1-e**2)/((1-e**2*np.sin(lat)**2)**(3/2))

    @property
    def first_eccentricity_squared(self) -> float:
        """
        First Eccentricity Squared :math:`e^2` of the ellipsoid, computed as:

        .. math::
            e^2 = 2f - f^2

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.first_eccentricity_squared
        0.0
        """
        return 2*self.f - self.f**2

    @property
    def second_eccentricity_squared(self) -> float:
        """
        Second Eccentricity Squared :math:`e'^2`, computed as:

        .. math::
            e'^2 = \\frac{a^2-b^2}{b^2}

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.second_eccentricity_squared
        0.0
        """
        return (self.a**2-self.b**2)/self.b**2

    @property
    def linear_eccentricity(self) -> float:
        """
        Linear Eccentricity :math:`E`, computed as:

        .. math::
            E = \\sqrt{a^2-b^2}

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.linear_eccentricity
        0.0
        """
        return np.sqrt(self.a**2-self.b**2)

    @property
    def aspect_ratio(self) -> float:
        """
        Aspect Ratio :math:`AR`, computed as:

        .. math::
            AR = \\frac{b}{a}

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.aspect_ratio
        1.0
        """
        return self.b/self.a

    @property
    def curvature_polar_radius(self) -> float:
        """
        Polar Radius of Curvature :math:`R_P`, computed as:

        .. math::
            \\begin{array}{ll}
            R_P &= \\frac{a^2}{b} = \\frac{a}{\\sqrt{1-e^2}} \\\\
            &= \\frac{a}{1-f}
            \\end{array}

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.curvature_polar_radius
        1.0
        """
        return self.a/(1-self.f)

    @property
    def arithmetic_mean_radius(self) -> float:
        """
        Mean Radius :math:`R_1` of the Three Semi-Axes, computed as:

        .. math::
            R_1 = a\\Big(1-\\frac{f}{3}\\Big)

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.arithmetic_mean_radius
        1.0
        """
        return self.a*(1-self.f/3)

    @property
    def authalic_sphere_radius(self) -> float:
        """
        Radius :math:`R_2` of a Sphere of Equal Area, computed as:

        .. math::
            R_2 = R_P \\Big(1-\\frac{2}{3}e'^2 + \\frac{26}{45}e'^4 - \\frac{100}{189}e'^6 + \\frac{7034}{14175}e'^8 - \\frac{220652}{467775}e'^{10} \\Big)

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.authalic_sphere_radius
        1.0
        """
        r = self.curvature_polar_radius
        es = np.sqrt(self.second_eccentricity_squared)
        return r*(1 - 2*es**2/3 + 26*es**4/45 - 100*es**6/189 + 7034*es**8/14175 - 220652*es**10/467775)

    @property
    def equivolumetric_sphere_radius(self) -> float:
        """
        Radius :math:`R_3` of a Sphere of Equal Volume, computed as:

        .. math::
            \\begin{array}{ll}
            R_3 &= \\sqrt[3]{a^2b} \\\\
            &= a\\sqrt[3]{1-f}
            \\end{array}

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.equivolumetric_sphere_radius
        1.0
        """
        return self.a*np.cbrt(1-self.f)

    @property
    def normal_gravity_constant(self) -> float:
        """
        Normal Gravity Formula Constant :math:`m`, computed as:

        .. math::
            m = \\frac{\\omega^2a^2b}{GM}

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.normal_gravity_constant
        14982844642.8839
        """
        return self.w**2*self.a**2*self.b/self.gm

    @property
    def dynamical_form_factor(self) -> float:
        """
        WGS 84 Dynamical Form Factor :math:`J_2`, computed as:

        .. math::
            J_2 = \\frac{e^2}{3} \\Big(1-\\frac{2me'}{15q_0}\\Big)

        where:

        .. math::
            q_0 = \\frac{1}{2}\\Big[\\Big(1+\\frac{3}{e'^2}\\Big)\\arctan e' - \\frac{3}{e'}\\Big]

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.dynamical_form_factor
        0.0
        """
        m = self.normal_gravity_constant
        e2 = self.first_eccentricity_squared
        es = np.sqrt(self.second_eccentricity_squared)
        if es == 0:
            return 0.0
        q0 = 0.5*((1+3/es**2)*np.arctan(es) - 3/es)
        return e2*(1-2*m*es/(15*q0))/3

    @property
    def second_degree_zonal_harmonic(self) -> float:
        """
        WGS 84 Second Degree Zonal Harmonic :math:`C_{2,0}`, computed as:

        .. math::
            C_{2,0} = -\\frac{J_2}{\\sqrt{5}}

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.second_degree_zonal_harmonic
        0.0
        """
        return -self.dynamical_form_factor/np.sqrt(5.0)

    @property
    def normal_gravity_potential(self) -> float:
        """
        Normal Gravity Potential :math:`U_0` of the WGS 84 Ellipsoid, computed
        as:

        .. math::
            U_0 = \\frac{GM}{E} \\arctan e' + \\frac{\\omega^2a^2}{3}

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.normal_gravity_potential
        6.6743e-11
        """
        es = np.sqrt(self.second_eccentricity_squared)
        if es == 0:
            return self.gm
        return self.gm*np.arctan(es)/self.linear_eccentricity + self.w**2*self.a**2/3

    @property
    def equatorial_normal_gravity(self) -> float:
        """
        Normal Gravity :math:`\\mathrm{g}_e` at the Equator, in
        :math:`\\frac{\\mathrm{m}}{\\mathrm{s}^2}`, computed as:

        .. math::
            \\mathrm{g}_e = \\frac{GM}{ab}\\Big(1-m-\\frac{me'q_0'}{6q_0}\\Big)

        where:

        .. math::
            \\begin{array}{ll}
            q_0 &= \\frac{1}{2}\\Big[\\Big(1+\\frac{3}{e'^2}\\Big)\\arctan e' - \\frac{3}{e'}\\Big] \\\\
            q_0' &= 3\\Big[\\Big(1+\\frac{1}{e'^2}\\Big)\\Big(1-\\frac{1}{e'}\\arctan e'\\Big)\\Big] - 1
            \\end{array}

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.equatorial_normal_gravity
        14982844642.8839
        """
        m = self.normal_gravity_constant
        es = np.sqrt(self.second_eccentricity_squared)
        if es == 0:
            return m
        q0 = 0.5*((1 + 3/es**2)*np.arctan(es) - 3/es)
        q0s = 3*((1 + 1/es**2)*(1 - np.arctan(es)/es)) - 1
        return self.gm * (1 - m - m*es*q0s/(6*q0))/(self.a*self.b)

    @property
    def polar_normal_gravity(self) -> float:
        """
        Normal Gravity :math:`\\mathrm{g}_p` at the Pole, in
        :math:`\\frac{\\mathrm{m}}{\\mathrm{s}^2}`, computed as:

        .. math::
            \\mathrm{g}_p = \\frac{GM}{a^2}\\Big(1+\\frac{me'q_0'}{3q_0}\\Big)

        where:

        .. math::
            \\begin{array}{ll}
            q_0 &= \\frac{1}{2}\\Big[\\Big(1+\\frac{3}{e'^2}\\Big)\\arctan e' - \\frac{3}{e'}\\Big] \\\\
            q_0' &= 3\\Big[\\Big(1+\\frac{1}{e'^2}\\Big)\\Big(1-\\frac{1}{e'}\\arctan e'\\Big)\\Big] - 1
            \\end{array}

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.polar_normal_gravity
        14982844642.8839
        """
        m = self.normal_gravity_constant
        es = np.sqrt(self.second_eccentricity_squared)
        if es == 0:
            return m
        q0 = 0.5*((1 + 3/es**2)*np.arctan(es) - 3/es)
        q0s = 3*((1 + 1/es**2)*(1 - np.arctan(es)/es)) - 1
        return self.gm * (1 + m*es*q0s/(3*q0))/self.a**2

    @property
    def mean_normal_gravity(self) -> float:
        """
        Mean Value :math:`\\bar{\\mathrm{g}}` of Normal Gravity, in
        :math:`\\frac{\\mathrm{m}}{\\mathrm{s}^2}`, computed as:

        .. math::
            \\bar{\\mathrm{g}} = \\mathrm{g}_e\\Big(1 + \\frac{1}{6}e^2 + \\frac{1}{3}k + \\frac{59}{360}e^4 + \\frac{5}{18}e^2k + \\frac{2371}{15120}e^6 + \\frac{259}{1080}e^4k + \\frac{270229}{1814400}e^8 + \\frac{9623}{45360}e^6k \\Big)

        where:

        .. math::
            k = \\frac{b\\mathrm{g}_p - a\\mathrm{g}_e}{a\\mathrm{g}_e} = \\frac{b\\mathrm{g}_p}{a\\mathrm{g}_e}-1

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.mean_normal_gravity
        14982844642.8839
        """
        e = np.sqrt(self.first_eccentricity_squared)
        gp = self.polar_normal_gravity
        ge = self.equatorial_normal_gravity
        k = (self.b*gp)/(self.a*ge)-1
        g = ge * (1 + e**2/6 + k/3 + 59*e**4/360 + 5*e**2*k/18 + 2371*e**6/15120 + 259*e**4*k/1080 + 270229*e**8/1814400 + 9623*e**6*k/45360)
        return g

    @property
    def mass(self) -> float:
        """
        The Mass :math:`M` of the ellipsoid, in kg, computed as:

        .. math::
            M = \\frac{GM}{G}

        where :math:`G` is the universal constant of gravitation equal to
        :math:`6.67428\\times 10^{-11} \\frac{\\mathrm{m}^3}{\\mathrm{kg} \\mathrm{s}^2}`

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.mass
        1.0
        """
        return self.gm/UNIVERSAL_GRAVITATION_CODATA2018

    @property
    def is_geodetic(self) -> bool:
        """
        Check whether the ellipsoid model is geodetic (describes planet Earth.)

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.is_geodetic
        False
        """
        is_geodetic = np.isclose(self.a, EARTH_EQUATOR_RADIUS)
        is_geodetic &= np.isclose(self.f, EARTH_FLATTENING)
        is_geodetic &= np.isclose(self.gm, EARTH_GM)
        is_geodetic &= np.isclose(self.w, EARTH_ROTATION)
        return is_geodetic

    @property
    def sidereal_day(self) -> float:
        """
        Sidereal Day, :math:`T_{sid}`, in seconds, computed as:

        .. math::
            T_{sid} = \\frac{2\\pi}{\\omega}

        Example
        -------
        >>> ref = ahrs.utils.ReferenceEllipsoid()
        >>> ref.sidereal_day
        6.283185307179586
        """
        return 2*np.pi/self.w



"""
Contains the core utilities for the proper use of AHRS: assertions, data
handling, etc.

These functions have no other goal in this package than to be used by other
modules. They are not meant to be used by the user.

This module is private. All functions and objects are available in the main
``ahrs`` namespace, or its corresponding submodule - use that instead.

Basic requirements to build assertions:

- ``assert`` functions should return either ``True`` or ``False``.
- If the assertion fails, an exception should be raised.
- The exception should have a meaningful message.
"""

import numpy as np

def _assert_list_of_strings(item, item_name: str = 'iterable'):
    """Assert it is a list of strings"""
    if not all(isinstance(i, str) for i in item):
        raise TypeError(f"{item_name} must be a list of strings.")

def _assert_valid_array_type(item, item_name: str = 'iterable'):
    """Assert it is a list, tuple, or numpy array"""
    # NOTE: This could be changed to a more pythonic solution looking for the
    # dunder method __iter__(), but that yields strings too.
    if not isinstance(item, (list, tuple, np.ndarray)):
        raise TypeError(f"{item_name} must be given as an array. Got {type(item)}")

def _assert_numerical_positive_variable(item, item_name: str = 'input'):
    if isinstance(item, bool):
        raise TypeError(f"Value '{item_name}' must be numeric.")
    if not isinstance(item, (int, float)):
        raise TypeError(f"Value '{item_name}' must be int or float.")
    if item <= 0.0:
        raise ValueError(f"Value '{item_name}' must be a non-zero number.")

def _assert_numerical_iterable(item, item_name: str = 'iterable'):
    """Assert it is a list, tuple, or numpy array, and that it has numerical values"""
    _assert_valid_array_type(item, item_name)
    item_copy = np.array(item)
    if not(item_copy.dtype == np.dtype(int) or item_copy.dtype == np.dtype(float)):
        raise TypeError(f"{item_name} must have numerical values. Got {item_copy.dtype.name}")

def _assert_same_shapes(item1, item2, item_names: list = None):
    for item in [item1, item2]:
        if not isinstance(item, (list, tuple, np.ndarray)):
            raise TypeError(f"{item} must be an array. Got {type(item)}")
    if item_names is None:
        item_names = ['item1', 'item2']
    item1, item2 = np.array(item1), np.array(item2)
    if item1.shape != item2.shape:
        raise ValueError(f"{item_names[0]} and {item_names[1]} must have the same shape. Got {item1.shape} and {item2.shape}")

def _assert_acc_mag_inputs(acc, mag):
    _assert_numerical_iterable(acc, 'acc')
    _assert_numerical_iterable(mag, 'mag')
    acc = np.array(acc)
    mag = np.array(mag)
    if acc.shape != mag.shape:
        raise ValueError("acc and mag are not the same size")

def get_nan_intervals(data: np.ndarray) -> list:
    """
    Get indices of NaN samples in data array.

    Based on an answer by `JonSG
    <https://codereview.stackexchange.com/users/105468/jonsg>`_ in
    `StackExchange <https://codereview.stackexchange.com/a/262803>`_.

    Parameters
    ----------
    data : array
        Data array.

    Returns
    -------
    intervals : list
        List of intervals of data samples with NaN.

    Examples
    --------
    >>> A = np.random.random((10, 3))
    >>> A[[1, 3, 4, 5, 8, 9]] = np.nan
    >>> A
    array([[0.74178186, 0.28444646, 0.35219214],
           [       nan,        nan,        nan],
           [0.05090735, 0.04161279, 0.10590561],
           [       nan,        nan,        nan],
           [       nan,        nan,        nan],
           [       nan,        nan,        nan],
           [0.31920402, 0.1523402 , 0.18907205],
           [0.32899368, 0.20042986, 0.73725579],
           [       nan,        nan,        nan],
           [       nan,        nan,        nan]])
    >>> get_nan_intervals(A)
    [(1, 1), (3, 5), (8, 9)]
    """
    data = np.copy(data)
    if data.ndim not in [1, 2]:
        raise ValueError(f"data array must be 1- or 2-dimensional. It has {data.ndim} dimensions.")
    isnan_list = np.any(np.isnan(data), axis=1) if data.ndim > 1 else np.isnan(data)
    nan_indices = np.where(isnan_list)[0]
    intervals = np.split(nan_indices, np.where(np.diff(nan_indices) > 1)[0] + 1)
    if len(intervals) == 0:
        return []
    return [(interval[0], interval[-1]) for interval in intervals]



# -*- coding: utf-8 -*-
"""
Attitude Estimators
===================

These are the most common attitude filters.

"""

from .angular import AngularRate
from .aqua import AQUA
from .complementary import Complementary
from .davenport import Davenport
from .ekf import EKF
from .famc import FAMC
from .flae import FLAE
from .fourati import Fourati
from .fqa import FQA
from .tilt import Tilt
from .madgwick import Madgwick
from .mahony import Mahony
from .oleq import OLEQ
from .quest import QUEST
from .roleq import ROLEQ
from .saam import SAAM
from .triad import TRIAD
from .fkf import FKF
from .ukf import UKF


# -*- coding: utf-8 -*-
"""
These are tools that are not specific to AHRS algorithms, but are used by them.

"""

from .metrics import *
from .wgs84 import *
from .wmm import *
from .geodesy import ReferenceEllipsoid



# -*- coding: utf-8 -*-
"""
Considering an attitude determination model from a pair of vector observations:

.. math::
    \\mathbf{D}^b = \\mathbf{CD}^r

where :math:`\\mathbf{D}_i^b=\\begin{bmatrix}D_{x,i}^b & D_{y,i}^b & D_{z,i}^b\\end{bmatrix}^T`
and :math:`\\mathbf{D}_i^r=\\begin{bmatrix}D_{x,i}^r & D_{y,i}^r & D_{z,i}^r\\end{bmatrix}^T`
are the *i*-th pair of normalized vector observations from the body frame :math:`b`
and the reference frame :math:`r`.

The goal is to find the optimal attitude :math:`\\mathbf{C}\\in\\mathbb{R}^{3\\times 3}`
relating both vectors. The famous `Wahba's problem
<https://en.wikipedia.org/wiki/Wahba%27s_problem>`_ can help us to find
:math:`\\mathbf{C}` from a set of observations and a least-squares method of
the form:

.. math::
    L(\\mathbf{C}) = \\sum_{i=1}^n a_i \\|\\mathbf{D}_i^b - \\mathbf{CD}_i^r \\|^2

being :math:`a_i` the weight of the *i*-th sensor output. The goal of **OLEQ**
is to find this optimal attitude, but in the form of a quaternion
:cite:p:`zhou2018`.

First, notice that the attitude matrix is related to quaternion
:math:`\\mathbf{q}=\\begin{bmatrix}q_w & q_x & q_y & q_z\\end{bmatrix}^T` via:

.. math::
    \\mathbf{C} = \\begin{bmatrix}\\mathbf{P}_1\\mathbf{q} & \\mathbf{P}_2\\mathbf{q} & \\mathbf{P}_3\\mathbf{q}\\end{bmatrix}

where the decomposition matrices are:

.. math::
    \\begin{array}{rcl}
    \\mathbf{P}_1 &=&
    \\begin{bmatrix}q_w & q_x & -q_y & -q_z \\\\ -q_z & q_y & q_x & -q_w \\\\ q_y & q_z & q_w & q_x \\end{bmatrix} \\\\
    \\mathbf{P}_2 &=&
    \\begin{bmatrix}q_z & q_y & q_x & q_w \\\\ q_w & -q_x & q_y & -q_z \\\\ -q_x & -q_w & q_z & q_y \\end{bmatrix} \\\\
    \\mathbf{P}_3 &=&
    \\begin{bmatrix}-q_y & q_z & -q_w & q_x \\\\ q_x & q_w & q_z & q_y \\\\ q_w & -q_x & -q_y & q_z \\end{bmatrix}
    \\end{array}

It is accepted that :math:`\\mathbf{P}_1^T=\\mathbf{P}_1^\\dagger`,
:math:`\\mathbf{P}_2^T=\\mathbf{P}_2^\\dagger`, and :math:`\\mathbf{P}_3^T=\\mathbf{P}_3^\\dagger`,
where the notation :math:`^\\dagger` stands for the `Moore-Penrose pseudo-
inverse <https://en.wikipedia.org/wiki/Moore%E2%80%93Penrose_inverse>`_. So,
the reference and observation vectors can be related to the quaternion with a
:math:`4\\times 4` matrix of the form:

.. math::
    \\begin{array}{rcl}
    \\mathbf{D}^b &=& \\mathbf{K}(\\mathbf{q}) \\mathbf{q} \\\\
    \\mathbf{D}^b &=& \\big(D_x^r\\mathbf{P}_1 + D_y^r\\mathbf{P}_2 + D_z^r\\mathbf{P}_3\\big) \\mathbf{q}
    \\end{array}

Knowing that :math:`\\mathbf{K}^T(\\mathbf{q})=\\mathbf{K}^\\dagger(\\mathbf{q})`,
the expression can be expanded to:

.. math::
    \\begin{array}{rcl}
    \\mathbf{K}^T(\\mathbf{q})\\mathbf{D}^b &=&
    D_x^r\\mathbf{P}_1^T\\mathbf{D}^b + D_y^r\\mathbf{P}_2^T\\mathbf{D}^b + D_z^r\\mathbf{P}_3^T\\mathbf{D}^b \\\\
    \\mathbf{Wq} &=& D_x^r\\mathbf{M}_1\\mathbf{q} + D_y^r\\mathbf{M}_2\\mathbf{q} + D_z^r\\mathbf{M}_3\\mathbf{q}
    \\end{array}

where :math:`\\mathbf{W}` is built with:

.. math::
    \\begin{array}{rcl}
    \\mathbf{W} &=& D_x^r\\mathbf{M}_1 + D_y^r\\mathbf{M}_2 + D_z^r\\mathbf{M}_3 \\\\ && \\\\
    \\mathbf{M}_1 &=&
    \\begin{bmatrix}
    D_x^b & 0 & D_z^b & -D_y^b \\\\
    0 & D_x^b & D_y^b & D_z^b \\\\
    D_z^b & D_y^b & -D_x^b & 0 \\\\
    -D_y^b & D_z^b & 0 & -D_x^b
    \\end{bmatrix} \\\\
    \\mathbf{M}_2 &=&
    \\begin{bmatrix}
    D_y^b & -D_z^b & 0 & D_x^b \\\\
    -D_z^b & -D_y^b & D_x^b & 0 \\\\
    0 & D_x^b & D_y^b & D_z^b \\\\
    D_x^b & 0 & D_z^b & -D_y^b
    \\end{bmatrix} \\\\
    \\mathbf{M}_3 &=&
    \\begin{bmatrix}
    D_z^b & D_y^b & -D_x^b & 0 \\\\
    D_y^b & -D_z^b & 0 & D_x^b \\\\
    -D_x^b & 0 & -D_z^b & D_y^b \\\\
    0 & D_x^b & D_y^b & D_z^b
    \\end{bmatrix}
    \\end{array}

Now the attitude estimation is shifted to :math:`\\mathbf{Wq}=\\mathbf{q}`. If
treated as an iterative dynamical system, the quatenion at the *n*-th itreation
is calculated as:

.. math::
    \\mathbf{q}(n) = \\mathbf{Wq}(n-1)

It is possible to list all rotation equations as:

.. math::
    \\begin{bmatrix}
    \\sqrt{a_1}\\mathbf{I}_4 \\\\ \\vdots \\\\ \\sqrt{a_n}\\mathbf{I}_4
    \\end{bmatrix} \\mathbf{q} =
    \\begin{bmatrix}
    \\sqrt{a_1}\\mathbf{W}_1 \\\\ \\vdots \\\\ \\sqrt{a_n}\\mathbf{W}_n
    \\end{bmatrix} \\mathbf{q}

Leading to a pre-multiplication of the form:

.. math::
    \\mathbf{q} = \\Big(\\sum_{i=1}^na_i\\mathbf{W}_i\\Big)\\mathbf{q}

A stable and continuous solution to each equation is done by pre-multiplying
:math:`\\frac{1}{2}(\\mathbf{W}_i+\\mathbf{I}_4)`.

.. math::
    \\begin{bmatrix}
    \\sqrt{a_1}\\mathbf{I}_4 \\\\ \\vdots \\\\ \\sqrt{a_n}\\mathbf{I}_4
    \\end{bmatrix} \\mathbf{q} =
    \\begin{bmatrix}
    \\frac{1}{2}\\sqrt{a_1}(\\mathbf{W}_1+\\mathbf{I}_4) \\\\ \\vdots \\\\ \\frac{1}{2}\\sqrt{a_n}(\\mathbf{W}_n+\\mathbf{I}_4)
    \\end{bmatrix} \\mathbf{q}

Based on `Brouwer's fixed-point theorem <https://en.wikipedia.org/wiki/Brouwer_fixed-point_theorem>`_,
it is possible to recursively obtain the normalized optimal quaternion by
rotating a randomly given initial quaternion, :math:`\\mathbf{q}_\\mathrm{rand}`,
over and over again indefinitely.

.. math::
    \\mathbf{q} = \\frac{\\mathbf{W} + \\mathbf{I}}{2} \\mathbf{q}_\\mathrm{rand}

This equals the least-square of the set of pre-computed single rotated
quaternions.

"""

import numpy as np
from ..common.mathfuncs import cosd
from ..common.mathfuncs import sind
from ..utils.core import _assert_numerical_iterable

class OLEQ:
    """
    Optimal Linear Estimator of Quaternion

    Parameters
    ----------
    acc : numpy.ndarray, default: None
        N-by-3 array with measurements of acceleration in in m/s^2
    mag : numpy.ndarray, default: None
        N-by-3 array with measurements of magnetic field in mT
    weights : numpy.ndarray, default: ``[1., 1.]``
        Array with weights for each sensor measurement. The first item weights
        the observed acceleration, while second item weights the observed
        magnetic field.
    magnetic_ref : float or numpy.ndarray
        Local magnetic reference.
    frame : str, default: ``'NED'``
        Local tangent plane coordinate frame. Valid options are right-handed
        ``'NED'`` for North-East-Down and ``'ENU'`` for East-North-Up.

    Raises
    ------
    ValueError
        When dimension of input arrays ``acc`` and ``mag`` are not equal.

    Examples
    --------
    >>> acc_data.shape, mag_data.shape      # NumPy arrays with sensor data
    ((1000, 3), (1000, 3))
    >>> from ahrs.filters import OLEQ
    >>> orientation = OLEQ(acc=acc_data, mag=mag_data)
    >>> orientation.Q.shape                 # Estimated attitude
    (1000, 4)

    """
    def __init__(self,
        acc: np.ndarray = None,
        mag: np.ndarray = None,
        weights: np.ndarray = None,
        magnetic_ref: np.ndarray = None,
        frame: str = 'NED'
        ):
        self.acc: np.ndarray = acc
        self.mag: np.ndarray = mag
        self.a: np.ndarray = weights if weights is not None else np.ones(2)
        self.frame: str = frame
        # Reference measurements
        self._set_reference_frames(magnetic_ref, self.frame)
        self._assert_validity_of_inputs()
        if self.acc is not None and self.mag is not None:
            self.Q = self._compute_all()

    def _set_reference_frames(self, mref: float, frame: str = 'NED') -> None:
        if not isinstance(frame, str):
            raise TypeError(f"'frame' must be a string. Got {type(frame)}.")
        if frame.upper() not in ['NED', 'ENU']:
            raise ValueError(f"Invalid frame '{frame}'. Try 'NED' or 'ENU'")
        #### Magnetic Reference Vector ####
        if mref is None:
            # Local magnetic reference of Munich, Germany
            from ..common.constants import MUNICH_LATITUDE, MUNICH_LONGITUDE, MUNICH_HEIGHT
            from ..utils.wmm import WMM
            wmm = WMM(latitude=MUNICH_LATITUDE, longitude=MUNICH_LONGITUDE, height=MUNICH_HEIGHT)
            cd, sd = cosd(wmm.I), sind(wmm.I)
            self.m_ref = np.array([sd, 0.0, cd]) if frame.upper() == 'NED' else np.array([0.0, cd, -sd])
        elif isinstance(mref, bool):
            raise TypeError(f"'mref' must be a float or numpy.ndarray. Got {type(mref)}.")
        elif isinstance(mref, (int, float)):
            # Use given magnetic dip angle (in degrees)
            cd, sd = cosd(mref), sind(mref)
            self.m_ref = np.array([sd, 0.0, cd]) if frame.upper() == 'NED' else np.array([0.0, cd, -sd])
        elif isinstance(mref, (list, tuple, np.ndarray)):
            # Magnetic reference is given as a vector
            self.m_ref = np.copy(mref)
        else:
            raise TypeError(f"Magnetic reference must be float, int, list, tuple or numpy.ndarray. Got{type(mref)}.")
        if self.m_ref.shape != (3,):
            raise ValueError(f"Magnetic reference vector must be of shape (3,). Got {self.m_ref.shape}.")
        if not any(self.m_ref):
            raise ValueError("Magnetic reference vector must contain non-zero values.")
        self.m_ref /= np.linalg.norm(self.m_ref)
        #### Gravitational Reference Vector ####
        self.a_ref = np.array([0.0, 0.0, -1.0]) if frame.upper() == 'NED' else np.array([0.0, 0.0, 1.0])

    def _assert_validity_of_inputs(self):
        """Asserts the validity of the inputs."""
        # Assert arrays
        for item in ['acc', 'mag', 'a', 'm_ref', 'a_ref']:
            if self.__getattribute__(item) is not None:
                if isinstance(self.__getattribute__(item), bool):
                    raise TypeError(f"Parameter '{item}' must be an array of numeric values.")
                _assert_numerical_iterable(self.__getattribute__(item), item)
                self.__setattr__(item, np.copy(self.__getattribute__(item)))
        if self.acc is not None and self.mag is None:
            raise ValueError("If 'acc' is given, 'mag' must also be given.")
        if self.mag is not None and self.acc is None:
            raise ValueError("If 'mag' is given, 'acc' must also be given.")
        # Assert weights
        if self.a.shape != (2,):
            raise ValueError(f"Dimension of 'weights' must be (2,). Got {self.a.shape}.")
        for item in self.a:
            if not isinstance(item, (int, float)):
                raise TypeError(f"'weights' must be an array of numeric values. Got {type(item)}.")
            if item < 0.0:
                raise ValueError(f"'weights' must be non-negative. Got {item}.")
        if not any(self.a > 0):
            raise ValueError("'weights' must contain positive values.")

    def _compute_all(self) -> np.ndarray:
        """
        Estimate the quaternions given all data.

        Attributes ``acc`` and ``mag`` must contain data.

        Returns
        -------
        Q : numpy.ndarray
            M-by-4 Array with all estimated quaternions, where M is the number
            of samples.

        """
        _assert_numerical_iterable(self.acc, 'Gravitational acceleration vector')
        _assert_numerical_iterable(self.mag, 'Geomagnetic field vector')
        if self.acc.shape != self.mag.shape:
            raise ValueError("acc and mag are not the same size")
        num_samples = np.atleast_2d(self.acc).shape[0]
        if num_samples < 2:
            return self.estimate(self.acc, self.mag)
        return np.array([self.estimate(self.acc[t], self.mag[t]) for t in range(num_samples)])

    def WW(self, Db: np.ndarray, Dr: np.ndarray) -> np.ndarray:
        """
        W Matrix

        .. math::
            \\mathbf{W} = D_x^r\\mathbf{M}_1 + D_y^r\\mathbf{M}_2 + D_z^r\\mathbf{M}_3

        Parameters
        ----------
        Db : numpy.ndarray
            Normalized tri-axial observations vector.
        Dr : numpy.ndarray
            Normalized tri-axial reference vector.

        Returns
        -------
        W_matrix : numpy.ndarray
            W Matrix.
        """
        bx, by, bz = Db
        Dx, Dy, Dz = Dr
        M1 = np.array([
            [bx, 0.0, bz, -by],
            [0.0, bx, by, bz],
            [bz, by, -bx, 0.0],
            [-by, bz, 0.0, -bx]])       # (eq. 18a)
        M2 = np.array([
            [by, -bz, 0.0, bx],
            [-bz, -by, bx, 0.0],
            [0.0, bx, by, bz],
            [bx, 0.0, bz, -by]])        # (eq. 18b)
        M3 = np.array([
            [bz, by, -bx, 0.0],
            [by, -bz, 0.0, bx],
            [-bx, 0.0, -bz, by],
            [0.0, bx, by, bz]])         # (eq. 18c)
        return Dx*M1 + Dy*M2 + Dz*M3    # (eq. 20)

    def estimate(self, acc: np.ndarray, mag: np.ndarray) -> np.ndarray:
        """
        Attitude Estimation

        Parameters
        ----------
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer.
        mag : numpy.ndarray
            Sample of tri-axial Magnetometer.

        Returns
        -------
        q : numpy.ndarray
            Estimated quaternion.

        """
        # Normalize measurements (eq. 1)
        a_norm = np.linalg.norm(acc)
        m_norm = np.linalg.norm(mag)
        if a_norm == 0 or m_norm == 0:    # handle NaN
            return None
        acc = np.copy(acc)/a_norm
        mag = np.copy(mag)/m_norm
        sum_aW = self.a[0]*self.WW(acc, self.a_ref) + self.a[1]*self.WW(mag, self.m_ref) # (eq. 31)
        R = 0.5*(np.identity(4) + sum_aW)       # (eq. 33)
        q = np.random.random(4)-0.5             # "random" quaternion (eq. 25)
        q /= np.linalg.norm(q)
        last_q = np.array([1., 0., 0., 0.])
        i = 0
        while np.linalg.norm(q-last_q) > 1e-8 and i <= 20:
            last_q = q
            q = R @ last_q                      # (eq. 24)
            q /= np.linalg.norm(q)
            i += 1
        return q/np.linalg.norm(q)



# -*- coding: utf-8 -*-
"""
A frame of reference specifies the position of an object in relation to a
reference within a coordinate system.

In the context of global navigation, the reference frame is used to define the
position of an object in relation to the Earth. The Earth is a non-inertial
frame, and the object is usually a sensor or a vehicle.

In this case, the sensor frame is usually attached to the object, and the
navigation frame is attached to the Earth.

There main frames are:

- **Earth-Centered Inertial Frame** (ECI), also noted as **i-frame**, has its
  origin at the center of mass of the Earth. The X-axis points towards the
  `vernal equinox <https://en.wikipedia.org/wiki/March_equinox>`_ in the
  equatorial plane. The Z-axis is along the rotation axis of the Earth. The
  Y-axis completes with a right-hand system. This is also sometimes known as
  the `Celestial Reference System <https://gssc.esa.int/navipedia/index.php/Conventional_Celestial_Reference_System)>`_
  (CRS.)
- **Earth-Centered Earth-Fixed Frame** (ECEF), also noted as **e-frame**, has
  its origin and Z-axis aligned to the i-frame, but rotates along with the
  Earth. Therefore, is Earth-Fixed.
- The **North-East-Down** (NED) system has its origin fixed at the center of
  gravity of the aircraft. It is defined relative to a specific location on
  Earth, described by latitude, longitude and altitude.
- The local **East-North-Up** (ENU) coordinates are formed from a plane tangent
  to the Earth's surface fixed to a specific location and hence it is sometimes
  known as a `Local Tangent <https://en.wikipedia.org/wiki/Local_tangent_plane_coordinates>`_
  or "Local Geodetic" plane.
- An **Azimuth-Elevation-Range** (AER) system uses the `spherical coordinates
  <https://en.wikipedia.org/wiki/Spherical_coordinate_system>`_ (az, elev,
  range) to represent position relative to a local origin. The local origin is
  described by the geodetic coordinates (latitude, longitude, height). Azimuth,
  elevation, and slant range are dependent on a local Cartesian system, for
  example, an ENU system.
"""

import numpy as np
from .constants import EARTH_EQUATOR_RADIUS
from .constants import EARTH_POLAR_RADIUS
from .constants import RAD2DEG
from .constants import DEG2RAD

def geodetic2ecef(lat: float, lon: float, h: float, a: float = EARTH_EQUATOR_RADIUS, b: float = EARTH_POLAR_RADIUS) -> np.ndarray:
    """
    Transform geodetic coordinates to Rectangular (Cartesian) Coordinates in
    the Earth-Centered Earth-Fixed frame.

    The cartesian coordinates of a point :math:`\\begin{pmatrix}x & y & z\\end{pmatrix}`
    can be calculated from the geodetic coordinates
    :math:`\\begin{pmatrix}\\phi & \\lambda & h\\end{pmatrix}` using the
    following equations :cite:p:`Wiki_GCC_geo2rect` :cite:p:`ESA_Coord_Conv`
    :cite:p:`noureldin2013`:

    .. math::

        \\begin{array}{rcl}
        x & = & (N + h) \\cos\\phi \\cos\\lambda \\\\
        y & = & (N + h) \\cos\\phi \\sin\\lambda \\\\
        z & = & \\big(\\left(1 - e^2\\right)N + h\\big) \\sin\\phi
        \\end{array}

    where :math:`\\phi` is the latitude, :math:`\\lambda` is the longitude,
    :math:`h` is the height, and :math:`N` is the radius of curvature in the
    prime vertical at the given latitude :math:`\\phi`:

    .. math::

        r = \\frac{a}{\\sqrt{1 - e^2 \\sin^2\\phi}}

    The first eccentricity of the ellipsoid **squared**, :math:`e^2`, is
    defined using the equatorial radius of the ellipsoid :math:`a`, and the
    polar radius of the ellipsoid :math:`b`:

    .. math::

        e^2 = \\frac{a^2-b^2}{a^2}

    These values default to Earth's ellipsoid.

    Parameters
    ----------
    lat : float
        Latitude, in degrees.
    lon : float
        Longitude, in degrees.
    h : float
        Height above ellipsoidal surface, in meters.
    a : float, default: 6378137.0
        Ellipsoid's equatorial radius (semi-major axis), in meters. Defaults to
        Earth's.
    ecc : float, default: 8.1819190842622e-2
        Ellipsoid's first eccentricity. Defaults to Earth's.

    Returns
    -------
    X : numpy.ndarray
        ECEF cartesian coordinates.
    """
    if abs(lat) > 90.0:
        raise ValueError(f"Latitude must be between -90 and 90 degrees. Got {lat}")
    if abs(lon) > 180.0:
        raise ValueError(f"Longitude must be between -180 and 180 degrees. Got {lon}")
    lat *= DEG2RAD
    lon *= DEG2RAD
    ecc2 = (a**2 - b**2)/a**2
    N = a/np.sqrt(1 - ecc2 *np.sin(lat)**2)
    X = np.zeros(3)
    X[0] = (N+h)*np.cos(lat)*np.cos(lon)
    X[1] = (N+h)*np.cos(lat)*np.sin(lon)
    X[2] = (N*(1.0-ecc2)+h)*np.sin(lat)
    return X

def geodetic2enu(lat: float, lon: float, h: float, lat0: float, lon0: float, h0: float, a: float = EARTH_EQUATOR_RADIUS, b: float = EARTH_POLAR_RADIUS) -> np.ndarray:
    """
    Transform geodetic coordinates to east-north-up (ENU) coordinates
    :cite:p:`noureldin2013`.

    Transform local geodetic coordinates :math:`\\begin{pmatrix}\\phi &
    \\lambda & h\\end{pmatrix}` to local East-North-Up (ENU) coordinates
    :math:`\\begin{pmatrix}x & y & z\\end{pmatrix}`.

    The origin of the local ENU frame has to be defined by the geodetic
    coordinates :math:`\\begin{pmatrix}\\phi_0 & \\lambda_0 & h_0\\end{pmatrix}`.

    The transformation is performed in two steps:

    1. Convert both geodetic coordinates to ECEF coordinates.

    .. math::

        \\begin{array}{rcl}
        x & = & (N + h) \\cos\\phi \\cos\\lambda \\\\
        y & = & (N + h) \\cos\\phi \\sin\\lambda \\\\
        z & = & \\big(\\left(1 - e^2\\right)N + h\\big) \\sin\\phi
        \\end{array}

    .. math::

        \\begin{array}{rcl}
        x_0 & = & (N_0 + h_0) \\cos\\phi_0 \\cos\\lambda_0 \\\\
        y_0 & = & (N_0 + h_0) \\cos\\phi_0 \\sin\\lambda_0 \\\\
        z_0 & = & \\big(\\left(1 - e^2\\right)N_0 + h_0\\big) \\sin\\phi_0
        \\end{array}

    2. Convert the difference between the two ECEF coordinates to ENU.

    where :math:`N_1` and :math:`N_2` are the radius of curvature in the prime
    vertical at the given latitude :math:`\\phi_1` and :math:`\\phi_2`,
    respectively, and :math:`e^2` is the square of the first eccentricity of
    the ellipsoid.

    The ENU coordinates are computed as follows:

    .. math::

        \\begin{array}{rcl}
        x_{\\mathrm{ENU}} & = & -\\sin\\lambda_0 \\, (x_0 - x) + \\cos\\lambda_0 \\, (y_0 - y) \\\\
        y_{\\mathrm{ENU}} & = & -\\sin\\phi_0 \\, \\cos\\lambda_0 \\, (x_0 - x) - \\sin\\phi_0 \\, \\sin\\lambda_0 \\, (y_0 - y) + \\cos\\phi_0 \\, (z_0 - z) \\\\
        z_{\\mathrm{ENU}} & = & \\cos\\phi_0 \\, \\cos\\lambda_0 \\, (x_0 - x) + \\cos\\phi_0 \\, \\sin\\lambda_0 \\, (y_0 - y) + \\sin\\phi_0 \\, (z_0 - z)
        \\end{array}

    Parameters
    ----------
    lat : float
        Latitude of local origin, in degrees.
    lon : float
        Longitude of local origin, in degrees.
    h : float
        Height above ellipsoidal surface of local origin, in meters.
    lat0 : float
        Latitude of point of interesr, in degrees.
    lon0 : float
        Longitude of point of interest, in degrees.
    h0 : float
        Height above ellipsoidal surface of point of interest, in meters.
    a : float, default: 6378137.0
        Ellipsoid's equatorial radius (semi-major axis), in meters. Defaults to
        Earth's.
    ecc : float, default: 8.1819190842622e-2
        Ellipsoid's first eccentricity. Defaults to Earth's.

    Returns
    -------
    enu : numpy.ndarray
        ENU cartesian coordinates [east, north, up].

    The ENU coordinates are computed as follows:
    """
    x1, y1, z1 = geodetic2ecef(lat, lon, h, a, b)
    x2, y2, z2 = geodetic2ecef(lat0, lon0, h0, a, b)
    return ecef2enuv(x1, y1, z1, x2, y2, z2, lat0, lon0)

def ecef2geodetic(x: float, y: float, z: float, a: float = EARTH_EQUATOR_RADIUS, b: float = EARTH_POLAR_RADIUS) -> np.ndarray:
    """
    Transform cartesian coordinates in ECEF-frame to Geodetic Coordinates
    :cite:p:`ESA_Coord_Conv`.

    Given the Cartesian coordinates :math:`\\begin{pmatrix}x & y & z\\end{pmatrix}`
    in the Earth-Centered Earth-Fixed (ECEF) frame, we can start by computing
    the geodetic longitude :math:`\\lambda`:

    .. math::

        \\lambda = \\mathrm{arctan2}\\left(x, \\,y\\right)

    Then, we iteratively compute the geodetic latitude :math:`\\phi` using the
    initial estimation:

    .. math::

        \\phi_0 = \\mathrm{arctan2}\\big(z, \\,(1-e^2)p \\big)

    with :math:`p = \\sqrt{x^2 + y^2}`.

    Now we iterate until the difference between two consecutive latitudes
    (:math:`\\phi_i` and :math:`\\phi_{i-1}`) is smaller than a given threshold
    :math:`\\delta`. Each iteration updates the values as follows:

    .. math::

        \\begin{array}{rcl}
        N & \\leftarrow & \\frac{a}{\\sqrt{1 - e^2 \\sin^2(\\phi_{i-1})}} \\\\
        \\phi_i & \\leftarrow & \\mathrm{arctan2}\\big(z+e^2N\\sin(\\phi_{i-1}), \\,p\\big)
        \\end{array}

    where :math:`N` is the radius of curvature in the vertical prime, and
    :math:`e^2` is the square of the first eccentricity of the ellipsoid.

    .. math::

        e^2 = \\frac{a^2-b^2}{a^2}

    The value of :math:`\\delta` is empirically found to perform well when set
    to :math:`10^{-8}` in this implementation.

    The altitude (height) :math:`h` is computed as:

    .. math::

        h = \\frac{p}{\\cos\\phi} - N

    .. note::

        The altitude :math:`h` has an accuracy up to 0.1 m (10 cm), which is
        enough for most applications.

    The latitude and longitude are returned in degrees. The altitude is
    returned in meters.

    Parameters
    ----------
    x : float
        ECEF x-coordinate, in meters.
    y : float
        ECEF y-coordinate, in meters.
    z : float
        ECEF z-coordinate, in meters.
    a : float, default: 6378137.0
        Ellipsoid's equatorial radius, in meters. Defaults to Earth's.
    b : float, default: 6356752.3142
        Ellipsoid's polar radius, in meters. Defaults to Earth's.

    Returns
    -------
    lla : numpy.ndarray
        Geodetic coordinates [latitude, longitude, altitude].

    Examples
    --------
    >>> from ahrs.common.frames import ecef2geodetic
    >>> x = 4_201_000
    >>> y = 172_460
    >>> z = 4_780_100
    >>> ecef2geodetic(x, y, z)
    array([48.85616162,  2.35079383, 67.37006803])
    """
    e2 = (a**2 - b**2)/a**2  # Square of the first eccentricity: 2*f - f^2 = e^2 = (a^2 - b^2)/a^2
    p = np.sqrt(x**2 + y**2)
    lon = np.arctan2(y, x)
    # Iteratively compute latitude
    delta = 1e-8
    lat_old = 0
    lat = np.arctan2(z, (1-e2)*p)
    while abs(lat_old - lat) > delta:
        sin_lat = np.sin(lat)
        N = a / np.sqrt(1 - e2 * sin_lat**2)    # Radius of curvature in the vertical prime
        lat_old = lat
        lat = np.arctan2(z + e2 * N * sin_lat, p)
    h = p / np.cos(lat) - N
    # Convert to degrees
    lat *= RAD2DEG
    lon *= RAD2DEG
    return np.array([lat, lon, h])

def llf2ecef(lat: float, lon: float) -> np.ndarray:
    """
    Transform coordinates from LLF to ECEF

    Parameters
    ----------
    lat : float
        Latitude.
    lon : float
        Longitude.

    Returns
    -------
    R : numpy.ndarray
        Rotation Matrix.
    """
    return np.array([
        [-np.sin(lat), -np.sin(lon)*np.cos(lat), np.cos(lon)*np.cos(lat)],
        [ np.cos(lat), -np.sin(lon)*np.sin(lat), np.cos(lon)*np.sin(lat)],
        [         0.0,              np.cos(lon),             np.sin(lon)]])

def ecef2llf(lat: float, lon: float) -> np.ndarray:
    """
    Transform coordinates from ECEF to LLF

    Parameters
    ----------
    lat : float
        Latitude.
    lon : float
        Longitude.

    Returns
    -------
    R : numpy.ndarray
        Rotation Matrix.
    """
    return np.array([
        [            -np.sin(lat),              np.cos(lat),         0.0],
        [-np.sin(lon)*np.cos(lat), -np.sin(lon)*np.sin(lat), np.cos(lon)],
        [ np.cos(lon)*np.cos(lat),  np.cos(lon)*np.sin(lat), np.sin(lon)]])

def ecef2lla(x: float, y: float, z: float, a: float = EARTH_EQUATOR_RADIUS, b: float = EARTH_POLAR_RADIUS) -> np.ndarray:
    """Synonym of :func:`ecef2geodetic`."""
    return ecef2geodetic(x, y, z, a, b)

def eci2ecef(w: float, t: float = 0) -> np.ndarray:
    """
    Transformation between ECI and ECEF

    Parameters
    ----------
    w : float
        Rotation rate in rad/s
    t : float, default: 0.0
        Time since reference epoch.
    """
    return np.array([
        [ np.cos(w)*t, np.sin(w)*t, 0.0],
        [-np.sin(w)*t, np.cos(w)*t, 0.0],
        [         0.0,         0.0, 1.0]])

def ecef2enu(x: float, y: float, z: float, lat: float, lon: float, h: float, a: float = EARTH_EQUATOR_RADIUS, b: float = EARTH_POLAR_RADIUS) -> np.ndarray:
    """
    Transform geocentric XYZ coordinates in ECEF-frame to Local East-North-Up
    (ENU) cartesian coordinates :cite:p:`noureldin2013`.

    Parameters
    ----------
    x : float
        ECEF x-coordinate, in meters.
    y : float
        ECEF y-coordinate, in meters.
    z : float
        ECEF z-coordinate, in meters.
    lat : float
        Latitude, in degrees.
    lon : float
        Longitude, in degrees.
    h : float
        Height above ellipsoidal surface, in kilometers.
    a : float, default: 6378137.0
        Ellipsoid's equatorial radius (semi-major axis), in meters. Defaults to
        Earth's.
    ecc : float, default: 8.1819190842622e-2
        Ellipsoid's first eccentricity. Defaults to Earth's.

    Returns
    -------
    enu : numpy.ndarray
        ENU cartesian coordinates [east, north, up].

    Examples
    --------
    >>> from ahrs.common.frames import ecef2enu
    >>> x = 660_930.19276
    >>> y = -4_701_424.22296
    >>> z = 4_246_579.60463
    >>> lat = 42.0
    >>> lon = -82.0
    >>> h = 200.0
    >>> ecef2enu(x, y, z, lat, lon, h)
    array([186.27751933, 286.84222383, 939.69262095])
    """
    ecef = geodetic2ecef(lat, lon, h, a, b)
    x0, y0, z0 = ecef
    return ecef2enuv(x, y, z, x0, y0, z0, lat, lon)

def ecef2enuv(x: float, y: float, z: float, x0: float, y0: float, z0: float, lat: float, lon: float) -> np.ndarray:
    """
    Transform coordinates from ECEF to ENU

    To carry a transformation from the Earth-Centered Earth-Fixed (ECEF) frame
    to local coordinates we define a reference point :math:`\\mathbf{x}_r=
    \\begin{pmatrix}x_r & y_r & z_r\\end{pmatrix}`, and a point of interest
    :math:`\\mathbf{x}_p=\\begin{pmatrix}x_p & y_p & z_p\\end{pmatrix}`.

    We then transform from ECEF coordinates to the local navigation frame (LLF)
    using the rotation matrix :math:`R_{LLF}` :cite:p:`noureldin2013`
    :cite:p:`Wiki_Geographic_Conversions`:

    .. math::

        R_{LLF} = \\begin{bmatrix}
        -\\sin\\lambda_r & \\cos\\lambda_r & 0 \\\\
        -\\sin\\phi_r\\cos\\lambda_r & -\\sin\\phi_r\\sin\\lambda_r & \\cos\\phi_r \\\\
        \\cos\\phi_r\\cos\\lambda_r & \\cos\\phi_r\\sin\\lambda_r & \\sin\\phi_r
        \\end{bmatrix}

    The LLF-frame is referred to as **ENU** since its axes are aligned with the
    East, North and Up directions.

    The vector pointing from the reference point to the point of interest in
    the ENU frame is given by:

    .. math::

        \\begin{array}{rcl}
        \\mathbf{x}_{\\mathrm{ENU}} & = & R_{LLF} \\, \\mathbf{x}_{\\mathrm{ECEF}} \\\\
        \\begin{bmatrix}x \\\\ y \\\\ z\\end{bmatrix}_{\\mathrm{ENU}}
        & = &
        \\begin{bmatrix} -\\sin\\lambda_r & \\cos\\lambda_r & 0 \\\\
        -\\sin\\phi_r\\cos\\lambda_r & -\\sin\\phi_r\\sin\\lambda_r & \\cos\\phi_r \\\\
        \\cos\\phi_r\\cos\\lambda_r & \\cos\\phi_r\\sin\\lambda_r & \\sin\\phi_r
        \\end{bmatrix}
        \\begin{bmatrix}x_p - x_r \\\\ y_p - y_r \\\\ z_p - z_r\\end{bmatrix}
        \\end{array}

    The final ENU coordinates are:

    .. math::

        \\begin{array}{rcl}
        x_{\\mathrm{ENU}} & = & -\\sin\\lambda_r \\, (x_p - x_r) + \\cos\\lambda_r \\, (y_p - y_r) \\\\
        y_{\\mathrm{ENU}} & = & -\\sin\\phi_r \\, \\cos\\lambda_r \\, (x_p - x_r) - \\sin\\phi_r \\, \\sin\\lambda_r \\, (y_p - y_r) + \\cos\\phi_r \\, (z_p - z_r) \\\\
        z_{\\mathrm{ENU}} & = & \\cos\\phi_r \\, \\cos\\lambda_r \\, (x_p - x_r) + \\cos\\phi_r \\, \\sin\\lambda_r \\, (y_p - y_r) + \\sin\\phi_r \\, (z_p - z_r)
        \\end{array}

    Parameters
    ----------
    x : float
        ECEF x-coordinate, in meters.
    y : float
        ECEF y-coordinate, in meters.
    z : float
        ECEF z-coordinate, in meters.
    x0 : float
        ECEF x-coordinate of reference point, in meters.
    y0 : float
        ECEF y-coordinate of reference point, in meters.
    z0 : float
        ECEF z-coordinate of reference point, in meters.
    lat : float
        Latitude, in degrees.
    lon : float
        Longitude, in degrees.

    Returns
    -------
    enu : numpy.ndarray
        ENU cartesian coordinates [east, north, up].
    """
    lat *= DEG2RAD
    lon *= DEG2RAD
    u = x - x0
    v = y - y0
    w = z - z0
    t     =  np.cos(lon)*u + np.sin(lon)*v
    east  = -np.sin(lon)*u + np.cos(lon)*v
    up    =  np.cos(lat)*t + np.sin(lat)*w
    north = -np.sin(lat)*t + np.cos(lat)*w
    return np.array([east, north, up])

def enu2uvw(east: float, north: float, up: float, lat: float, lon: float, angle_unit: str = 'deg') -> np.ndarray:
    """
    UVW Mapping for ENU coordinates

    Parameters
    ----------
    east : float
        East.
    north : float
        North.
    up : float
        Up.
    lat : float
        Latitude.
    lon : float
        Longitude.
    angle_unit : str, default: 'deg'
        Unit of angles. Can be 'rad' or 'deg'.

    Returns
    -------
    uvw : numpy.ndarray
        UVW cartesian coordinates.
    """
    if angle_unit == 'deg':
        lat *= DEG2RAD
        lon *= DEG2RAD
    t = np.cos(lat) * up - np.sin(lat) * north
    w = np.sin(lat) * up + np.cos(lat) * north
    u = np.cos(lon) * t - np.sin(lon) * east
    v = np.sin(lon) * t + np.cos(lon) * east
    return np.array([u, v, w])

def enu2ecef(east: float, north: float, up: float, lat: float, lon: float, h: float, a: float = EARTH_EQUATOR_RADIUS, b: float = EARTH_POLAR_RADIUS) -> np.ndarray:
    """
    Transforms the local east-north-up (ENU) Cartesian coordinates specified by
    east, north, and up to the geocentric Earth-centered Earth-fixed (ECEF)
    Cartesian coordinates.

    Parameters
    ----------
    east : float
        East.
    north : float
        North.
    up : float
        Up.
    lat : float
        Latitude.
    lon : float
        Longitude.
    h : float
        Height above ellipsoidal surface, in kilometers.
    a : float, default: 6378137.0
        Ellipsoid's equatorial radius (semi-major axis), in meters. Defaults to
        Earth's.
    ecc : float, default: 8.1819190842622e-2
        Ellipsoid's first eccentricity. Defaults to Earth's.

    Returns
    -------
    ecef : numpy.ndarray
        ECEF cartesian coordinates.
    """
    ecef = geodetic2ecef(lat, lon, h, a, b)
    ## Rotating ENU to ECEF
    uvw = enu2uvw(east, north, up, lat, lon, 'deg')
    ## Origin + offset from origin equals position in ECEF
    return ecef + uvw

def _ltp_transformation(x: np.ndarray) -> np.ndarray:
    """
    Transform coordinates between NED and ENU.
    """
    x = np.copy(x)
    if x.shape[-1] != 3 or x.ndim > 2:
        raise ValueError(f"Given coordinates must have form (3, ) or (N, 3). Got {x.shape}")
    A = np.array([[0.0, 1.0, 0.0], [1.0, 0.0, 0.0], [0.0, 0.0, -1.0]])
    if x.ndim > 1:
        return (A @ x.T).T
    return A @ x

def ned2enu(x: np.ndarray) -> np.ndarray:
    """
    Transform coordinates from NED to ENU.

    Parameters
    ----------
    x : numpy.ndarray
        3D coordinates of point(s) to project.

    Returns
    -------
    x' : numpy.ndarray
        Transformed coordinates.
    """
    return _ltp_transformation(x)

def enu2ned(x: np.ndarray) -> np.ndarray:
    """
    Transform coordinates from ENU to NED.

    Parameters
    ----------
    x : numpy.ndarray
        3D coordinates of point(s) to project.

    Returns
    -------
    x' : numpy.ndarray
        Transformed coordinates.
    """
    return _ltp_transformation(x)

def aer2enu(az: float, elev: float, slant_range: float, deg: bool = True) -> np.ndarray:
    """
    Transform local azimuth-elevation-range (AER) spherical coordinates
    specified by ``az``, ``elev``, and ``slant_range`` to the local
    East-North-Up (ENU) Cartesian coordinates :cite:p:`strickland2020`.

    Parameters
    ----------
    az : float
        Azimuth measured clockwise from North.
    elev : float
        Elevation with respect to the local East-North plane.
    slant_range : float
        Distance from local origin.
    deg : bool, default: True
        If True, angles are given in degrees. Otherwise, they are in radians.

    Returns
    --------
    enu : numpy.ndarray
        ENU cartesian coordinates [east, north, up].
    """
    if deg:
        az *= DEG2RAD
        elev *= DEG2RAD
    r = slant_range*np.cos(elev)
    return np.array([r*np.sin(az), r*np.cos(az), slant_range*np.sin(elev)])

def enu2aer(east: float, north: float, up: float, deg: bool = True) -> np.ndarray:
    """
    Transform the local east-north-up (ENU) Cartesian coordinates specified by
    ``east``, ``north``, and ``up`` to the local azimuth-elevation-range (AER)
    Spherical coordinates :cite:p:`strickland2020`.

    Parameters
    ----------
    east : float
        X-coordinate of a point in the local ENU system.
    north : float
        Y-coordinate of a point in the local ENU system.
    up : float
        Z-coordinate of a point in the local ENU system.
    deg : bool, default: True
        If True, angles are returned in degrees. Otherwise, they are in radians.

    Returns
    -------
    aer : numpy.ndarray
        AER spherical coordinates [azimuth, elevation, slant_range].
    """
    r = np.linalg.norm([east, north])
    slant_range = np.linalg.norm([r, up])
    elev = np.arctan2(up, r)
    az = np.arctan2(east, north) % (2*np.pi)
    if deg:
        az *= RAD2DEG
        elev *= RAD2DEG
    return np.array([az, elev, slant_range])

def enu2dca(east: float, north: float, up: float, angle: float, deg: bool = True) -> np.ndarray:
    """
    Transform the local east-north-up (ENU) Cartesian coordinates specified by
    ``east``, ``north``, and ``up`` to the down-cross-above (DCA) reference
    frame :cite:p:`strickland2020`.

    The conversion is given by:

    .. math::

        \\begin{bmatrix}d \\\\ c \\\\ a\\end{bmatrix} =
        \\begin{bmatrix}
        \\sin\\theta & \\cos\\theta & 0 \\\\
        -\\cos\\theta & \\sin\\theta & 0 \\\\
        0 & 0 & 1
        \\end{bmatrix}
        \\begin{bmatrix}e \\\\ n \\\\ u\\end{bmatrix}

    Parameters
    ----------
    east : float
        X-coordinate of a point in the local ENU system.
    north : float
        Y-coordinate of a point in the local ENU system.
    up : float
        Z-coordinate of a point in the local ENU system.
    angle : float
        Angle measured clockwise from North.
    deg : bool, default: True
        If True, angles are given in degrees. Otherwise, they are in radians.

    Returns
    -------
    dca : numpy.ndarray
        DCA Cartesian coordinates [d, c, a].
    """
    if deg:
        angle *= DEG2RAD
    d = np.sin(angle)*east + np.cos(angle)*north
    c = -np.cos(angle)*east + np.sin(angle)*north
    return np.array([d, c, up])

def dca2enu(down: float, cross: float, above: float, angle: float, deg: bool = True) -> np.ndarray:
    """
    Transform the down-cross-above (DCA) Cartesian coordinates specified by
    ``down``, ``cross``, and ``above`` to the local east-north-up (ENU) system
    :cite:p:`strickland2020`.

    The conversion is given by:

    .. math::

        \\begin{bmatrix}e \\\\ n \\\\ u\\end{bmatrix} =
        \\begin{bmatrix}
        \\sin\\theta & -\\cos\\theta & 0 \\\\
        \\cos\\theta & \\sin\\theta & 0 \\\\
        0 & 0 & 1
        \\end{bmatrix}
        \\begin{bmatrix}d \\\\ c \\\\ a\\end{bmatrix}

    Parameters
    ----------
    down : float
        X-coordinate of a point in the DCA system.
    cross : float
        Y-coordinate of a point in the DCA system.
    above : float
        Z-coordinate of a point in the DCA system.
    angle : float
        Angle measured clockwise from North.
    deg : bool, default: True
        If True, angles are given in degrees. Otherwise, they are in radians.

    Returns
    -------
    enu : numpy.ndarray
        ENU Cartesian coordinates [east, north, up].
    """
    if deg:
        angle *= DEG2RAD
    east = np.sin(angle)*down - np.cos(angle)*cross
    north = np.cos(angle)*down + np.sin(angle)*cross
    return np.array([east, north, above])
